import inspect
import os
import socket
import subprocess
import time
import warnings
from datetime import datetime
from pathlib import Path
from typing import Union

from ..base import GroupCollection, StatusJob, ls
from ..core import _check_input_gfbt, _gen_group_collection, is_namespace
from ..jobs import job_queue
from ..pyslk_exceptions import PySlkBadProcessError, PySlkException
from ..raw import recall2_raw, recall_raw
from ..utils import get_recall_job_id, get_slk_pid, is_search_id

__all__ = [
    "recall_dev",
    "recall_single",
]


def recall_single(
    resources: Union[
        Path, str, int, list[Path], list[str], list[int], set[Path], set[str], set[int]
    ],
    destination: Union[Path, str, None] = None,
    resource_ids: bool = False,
    search_id: bool = False,
    recursive: bool = False,
    preserve_path: bool = True,
) -> int:
    """Run a slk_helpers recall

    :param resources: multiple resource ids (set resource_ids=True) or paths OR one search id (set search_id=True)
    :type resources: str or Path or int or list[str] or list[Path] or list[int] or set[str] or set[Path] or set[int]
    :param destination: destination directory to check whether some of the needed files are already available and
        don't need to be recalled
    :type destination: str or Path
    :param resource_ids: consider input as 'resource' as resource id or resource ids
    :type resource_ids: bool
    :param search_id: consider input as 'resource' as search id
    :type search_id: bool
    :param recursive: use the -R flag to recall recursively, Default: False
    :type recursive: bool
    :param preserve_path: preserve original file path, Default: True
    :type preserve_path: bool
    """
    if destination is not None:
        destination = os.path.expandvars(os.path.expanduser(destination))
    output: subprocess.CompletedProcess = recall2_raw(
        resources=resources,
        destination=destination,
        resource_ids=resource_ids,
        search_id=search_id,
        recursive=recursive,
        preserve_path=preserve_path,
        verbose=False,
        double_verbose=False,
        return_type=2,
    )
    if output.returncode == 0:
        return int(output.stdout.decode("utf-8").rstrip())
    elif output.returncode == 1:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )
    else:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )


# TODO: test
def recall(
    resource: Union[str, int, Path, list[str], GroupCollection, None] = None,
    search_id: Union[str, int, None] = None,
    search_query: Union[str, None] = None,
    recursive: bool = False,
    n_parallel_recalls_soft_max: int = 5,
    proceed_on_error: bool = False,
    verbose: bool = False,
    **kwargs,
) -> GroupCollection:
    """

    kwargs:
    * n_parallel_recalls_hard_max
    * split_recalls
    * wait_time_check_running_jobs (in seconds)
    * max_tape_number_per_search
    * wait_until_finished

    :param resource: list of files or a namespaces with files that should be recalled.
    :type resource: ``str``, ``list``, ``Path``
    :param search_id: id of a search
    :type search_id: ``int``, ``str``
    :param search_query: a search query
    :type search_query: ``str``
    :param recursive: do recursive search in the namespaces
    :type recursive: ``bool``
    :param n_parallel_recalls_soft_max: number of parallel recalls to perform
    :type n_parallel_recalls_soft_max: ``int``
    :param proceed_on_error: do not throw an error if a recall job cannot be submitted
    :type proceed_on_error: ``bool``
    :param verbose: verbose mode
    :type verbose: ``bool``
    :param kwargs: keyword arguments
    :type verbose: ``dict``
    :return: a collection of all
    :rtype: ``GroupCollection``
    """
    raise NotImplementedError(
        f"pyslk.{inspect.stack()[0][3]}: this function is not finally implement yet. Please use 'pyslk.recall_raw' "
        + "or the development version of this function 'pyslk.recall_dev'. The latter function has to be considered"
        + "as unstable and subject to spontaneous changes."
    )


def recall_dev(
    resource: Union[str, int, Path, list[str], GroupCollection, None] = None,
    search_id: Union[str, int, None] = None,
    search_query: Union[str, None] = None,
    recursive: bool = False,
    n_parallel_recalls_soft_max: int = 5,
    proceed_on_error: bool = False,
    verbose: bool = False,
    **kwargs,
) -> GroupCollection:
    """

    kwargs:
    * n_parallel_recalls_hard_max
    * split_recalls
    * wait_time_check_running_jobs (in seconds)
    * max_tape_number_per_search
    * wait_until_finished

    :param resource: list of files or a namespaces with files that should be recalled.
    :type resource: ``str``, ``list``, ``Path``
    :param search_id: id of a search
    :type search_id: ``int``, ``str``
    :param search_query: a search query
    :type search_query: ``str``
    :param recursive: do recursive search in the namespaces
    :type recursive: ``bool``
    :param n_parallel_recalls_soft_max: number of parallel recalls to perform
    :type n_parallel_recalls_soft_max: ``int``
    :param proceed_on_error: do not throw an error if a recall job cannot be submitted
    :type proceed_on_error: ``bool``
    :param verbose: verbose mode
    :type verbose: ``bool``
    :param kwargs: keyword arguments
    :type verbose: ``dict``
    :return: a collection of all
    :rtype: ``GroupCollection``
    """
    # check if input values are OK
    if isinstance(resource, GroupCollection):
        _check_input_gfbt(
            inspect.stack()[0][3], None, search_id, search_query, recursive
        )
    else:
        _check_input_gfbt(
            inspect.stack()[0][3], resource, search_id, search_query, recursive
        )

    msg: str
    # some default values for keyword arguments
    default_n_parallel_recalls_hard_max = -1
    default_split_recalls = True
    default_wait_time_check_running_jobs = 60
    default_max_tape_number_per_search = -1
    default_wait_until_finished = False
    # 600 seconds; 10 minutes
    default_max_time_to_wait_for_running_jobs = 600
    default_max_time_to_wait_for_job_id = 20
    # get max wait time
    max_wait_time_total: int = kwargs.get(
        "max_time_to_wait_for_running_jobs",
        default_max_time_to_wait_for_running_jobs,
    )
    max_wait_time_job_id_total: int = kwargs.get(
        "max_time_to_wait_for_job_id",
        default_max_time_to_wait_for_job_id,
    )
    # get wait time
    wait_time_check_running_jobs: int = kwargs.get(
        "wait_time_check_running_jobs",
        default_wait_time_check_running_jobs,
    )
    # wait time of loops
    wait_time: int

    groups: GroupCollection()
    # get or generate file groups
    if isinstance(resource, GroupCollection):
        if verbose:
            print(
                "~~~ got an instance of GroupCollection as input. No need to generate need GroupCollection"
            )
        groups = resource
    else:
        if verbose:
            print(
                "~~~ no instance of GroupCollection as input. Generate new instance of GroupCollection from input"
            )
        groups = _gen_group_collection(
            resource,
            search_id,
            search_query,
            recursive,
            max_tape_number_per_search=kwargs.get(
                "max_tape_number_per_search", default_max_tape_number_per_search
            ),
            split_recalls=kwargs.get("split_recalls", default_split_recalls),
            verbose=verbose,
        )

    # dicts to store job ids in (contain JOB_IDs AND GROUPS)
    all_jobs = set()
    active_jobs = set()
    successful_jobs = set()
    failed_jobs = set()
    error_not_started_jobs = set()
    cached_groups = set()
    # we iterate over the search Ids
    for search_id in groups:
        if verbose:
            print(f"~~~ processing group with search id '{search_id}'")

        group = groups.get_group(search_id)
        # remove error flag in file group (if set from previous recalls
        group.set_recall_error(False)
        # job needs to be started:

        # update caching info of this 'group'
        group.updated_caching_info()
        # skip group if all files are cached or group has job id + non-aborted job
        if group.is_cached() or (
            group.has_job_id()
            and group.get_job_status() not in [StatusJob.ABORTED, StatusJob.COMPLETED]
        ):
            if verbose:
                print("      skip this group (every cached or still-running recall job")
            cached_groups.add(group)
        else:
            # check if a job is COMPLETE, although we are  here (then the job is completed but not all files are cached)
            if group.has_job_id() and group.get_job_status() == StatusJob.COMPLETED:
                # warn user
                warnings.warn(
                    f"pyslk.{inspect.stack()[0][3]}: file group with search id '{search_id}' has already been "
                    + f"successfully recalled by job with id '{group.get_job_id()}' but not all files are cached."
                )
            # We set this to -1 so that we run through the while loop at least once!
            n_parallel_recalls = -1
            wait_time = 0
            while (
                len(active_jobs) > n_parallel_recalls
                and wait_time < max_wait_time_total
            ):
                # re-calculate number of allow parallel jobs
                n_parallel_recalls = _n_parallel_recalls_max(
                    n_parallel_recalls_soft_max,
                    kwargs.get(
                        "n_parallel_recalls_hard_max",
                        default_n_parallel_recalls_hard_max,
                    ),
                )
                # check running jobs
                if verbose:
                    print("   ~~~ update status of running jobs")
                # FIRST: go through all active recalls and check which are finished
                for active_job in active_jobs:
                    status = active_job.get_job_status()
                    if status.is_finished():
                        # job has finished
                        # ... add to successful or failed jobs
                        if status.is_successful():
                            successful_jobs.add(active_job)
                        else:
                            failed_jobs.add(active_job)
                            # set error flag in file group
                            active_job.set_recall_error(True)
                        # ... remove from active jobs
                        active_jobs.remove(active_job)
                if verbose:
                    print("   ~3020~~ status of recall jobs (total numbers)")
                    print(
                        "      waiting (not submitted too StrongLink): "
                        + f"{len(groups) - len(all_jobs) - len(cached_groups)}"
                    )
                    print(f"      allowed: {n_parallel_recalls}")
                    print(f"      submitted (active/queued): {len(active_jobs)}")
                    print(f"      finished, successful: {len(successful_jobs)}")
                    print(f"      finished, failed: {len(failed_jobs)}")
                    print(
                        f"      failed to be submitted: {len(error_not_started_jobs)}"
                    )
                if len(active_jobs) > n_parallel_recalls:
                    # inform user
                    if verbose:
                        print(
                            f"   ~~~ cannot submit new jobs for now; sleep {wait_time_check_running_jobs} seconds "
                            + "until re-check jobs"
                        )
                    # sleep / wait
                    time.sleep(wait_time_check_running_jobs)
                    wait_time = wait_time + wait_time_check_running_jobs
            # SECOND: submit next job
            if verbose:
                print(f"   ~~~ submitting recall job for search id {search_id}")
            proc: subprocess.Popen = recall_raw(search_id, recursive, return_type=3)
            # count submitted job
            all_jobs.add(group)
            pid: int = -1
            try:
                # try to get pid
                pid = get_slk_pid(proc)
            except PySlkBadProcessError as e:
                # no pid available
                now = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
                msg = (
                    "slk recall did not start properly; could not grab process ID of recall with message "
                    + f"'{str(e)}'; please see slk-cli log for details (~/.slk/slk-cli.log) close to timestamp "
                    + f"'{now}' and with hostname '{socket.gethostname()}'"
                )
                if not proceed_on_error:
                    raise PySlkBadProcessError(f"pyslk.{inspect.stack()[0][3]}: {msg}")
                # set error flag in file group
                group.set_recall_error(True)
                # store problematic group:
                error_not_started_jobs.add(group)
                # warn user
                warnings.warn(f"pyslk.{inspect.stack()[0][3]}: {msg}")
                if verbose:
                    print(f"   ~~~ {msg}")
            # set pid if available (then we are here)
            if verbose:
                print(f"   ~~~ slk recall started with pid {pid}")
            group.put_process_id(pid)
            job_id: Union[int, None] = None
            wait_time = 0
            while job_id is None and wait_time < max_wait_time_job_id_total:
                job_id = get_recall_job_id(pid)
                time.sleep(2)
                wait_time = wait_time + 2
            if job_id is not None:
                if verbose:
                    print(
                        f"   ~~~ recall job submitted successfully with job id {job_id}"
                    )
                group.put_job_id(job_id)
                active_jobs.add(group)
            else:
                # no job_id available
                now = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
                msg = (
                    "recall did not start as expected; please see slk-cli log for details "
                    + f"(~/.slk/slk-cli.log); related log entries contain the pid '{pid}' in column four, the "
                    + f"hostname '{socket.gethostname()}' in column three and a time stamp similar to '{now}' "
                    + "in columns one and two"
                )
                if not proceed_on_error:
                    raise PySlkBadProcessError(f"pyslk.{inspect.stack()[0][3]}: {msg}")
                # set error flag in file group
                group.set_recall_error(True)
                # store problematic group:
                error_not_started_jobs.add(group)
                # raise warning
                warnings.warn(f"pyslk.{inspect.stack()[0][3]}: {msg}")
                # write verbose
                if verbose:
                    print(f"   ~~~ {msg}")

    if kwargs.get("wait_until_finished", default_wait_until_finished):
        job_status: StatusJob
        wait_time = 0
        while len(active_jobs) > 0 and wait_time < max_wait_time_total:
            # inform user
            if verbose:
                print(
                    f"   ~~~ wait for all jobs to be finished; {len(active_jobs)}/{len(all_jobs)} not finished yet; "
                    + f"waiting {wait_time_check_running_jobs} seconds before checking the status of running jobs the "
                    + "next time"
                )
            # sleep / wait
            time.sleep(wait_time_check_running_jobs)
            wait_time = wait_time + wait_time_check_running_jobs
            for group in active_jobs:
                job_status = group.get_job_status()
                if job_status.is_finished():
                    if job_status.is_successful():
                        successful_jobs.add(group)
                    else:
                        failed_jobs.add(group)
                    active_jobs.remove(group)

    if verbose:
        print("  ~~~ number of tape groups ~~~")
        print(f"      submitted as recall jobs: {len(all_jobs)} / {len(groups)}")
        print(f"      available in the cache: {len(cached_groups)} / {len(groups)}")
        print("  ~~~ number of recall jobs ~~~")
        print(f"      submitted (active/queued): {len(active_jobs)}")
        print(f"      finished, successful: {len(successful_jobs)}")
        print(f"      finished, failed: {len(failed_jobs)}")
        print(f"      failed to be submitted: {len(error_not_started_jobs)}")

    return groups


# TODO: test
def _n_parallel_recalls_max(
    n_parallel_recalls_soft_max: int, n_parallel_recalls_hard_max: int = -1
) -> int:
    if n_parallel_recalls_hard_max > 0:
        return n_parallel_recalls_hard_max
    job_queue_info = job_queue()
    sl_active_jobs = job_queue_info["raw"]["active"]
    sl_queued_jobs = job_queue_info["raw"]["queued"]
    basic_max: int = max(14 - sl_active_jobs, 0) + max(5 - sl_queued_jobs, 0)
    return max(basic_max, n_parallel_recalls_soft_max)


def construct_target_file_path_local(
    path_or_id: Union[Path, str, int],
    dest_dir: Union[str, Path],
    recursive: bool = False,
    preserve_path: bool = False,
) -> set:
    """Construct the target path of each file to be retrieved when the given parameters are set

    :param path_or_id: search id or gns path of resources to retrieve
    :type path_or_id: ``str`` or ``int``
    :param dest_dir: destination directory for retrieval
    :type dest_dir: ``str``
    :param recursive: use the -R flag to retrieve recursively, Default: False
    :type recursive: ``bool``
    :param preserve_path: preserve namespace in destination
    :type preserve_path: ``bool``
    :returns: set of target file paths
    :rtype: ``set``
    """
    if not is_search_id(path_or_id) and is_namespace(path_or_id) and not recursive:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: 'path_or_id' considered as a namespace/directory, but recursive is not set"
        )
    if preserve_path:
        files = {Path(dest_dir, f[1:]) for f in ls(path_or_id)["filename"]}
    elif is_search_id(path_or_id):
        files = {
            Path(dest_dir, f[1:]) for f in ls(path_or_id, full_path=False)["filename"]
        }
    else:
        files = {
            Path(dest_dir, Path(f).relative_to(Path(path_or_id).parent))
            for f in ls(path_or_id, full_path=False)["filename"]
        }
    return files



import inspect
import json
import subprocess
import typing

from ..base import StatusJob
from ..pyslk_exceptions import PySlkException
from ..raw import job_exists_raw, job_queue_raw, job_status_raw

__all__ = [
    "_is_job_completed",
    "job_exists",
    "get_job_status",
    "is_job_finished",
    "is_job_queued",
    "is_job_processing",
    "is_job_successful",
    "job_queue",
]


def _is_job_completed(job_id: int) -> typing.Optional[bool]:
    """Check the status of a tape

    :param job_id: id of a recall job in StrongLink
    :type job_id: ``int``
    :returns: True if job with job_id is finished and has status 'COMPLETED'; else False
    :rtype: ``bool`` or ``None``
    """
    job_status: StatusJob = get_job_status(job_id)
    if job_status is None:
        return None
    return job_status.is_completed()


def job_exists(job_id: int) -> bool:
    """Check if job exists

    :param job_id: id of a recall job in StrongLink
    :type job_id: ``int``
    :returns: True if job exists; False otherwise
    :rtype: ``bool``
    """
    if not isinstance(job_id, int):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: wrong input type; need 'int'; got '{type(job_id).__name__}'"
        )
    output: subprocess.CompletedProcess = job_exists_raw(job_id, return_type=2)
    if output.returncode == 0:
        return True
    elif output.returncode == 1:
        return False
    else:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )


def get_job_status(job_id: int) -> typing.Optional[StatusJob]:
    """Check the status of a job

    :param job_id: id of a recall job in StrongLink
    :type job_id: ``int``
    :returns: job status; None, if job_id does not exist
    :rtype: ``pyslk.StatusJob`` or ``None``
    """
    if not isinstance(job_id, int):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: wrong input type; need 'int'; got '{type(job_id).__name__}'"
        )
    if not job_exists(job_id):
        return None
    output: subprocess.CompletedProcess = job_status_raw(job_id, return_type=2)
    if output.returncode in [0, 1]:
        return StatusJob(output.stdout.decode("utf-8").rstrip())
    else:
        raise PySlkException(
            f"pyslk.{inspect.stack()[0][3]}: "
            + f"{output.stdout.decode('utf-8').rstrip()} "
            + f"{output.stderr.decode('utf-8').rstrip()}"
        )


def is_job_successful(job_id: int) -> typing.Optional[bool]:
    """Check the status of a tape

    :param job_id: id of a recall job in StrongLink
    :type job_id: ``int``
    :returns: True if job with job_id is finished and was successful; else False
    :rtype: ``bool`` or ``None``
    """
    job_status: StatusJob = get_job_status(job_id)
    if job_status is None:
        return None
    return job_status.is_successful()


def is_job_finished(job_id: int) -> typing.Optional[bool]:
    """Check the status of a tape

    :param job_id: id of a recall job in StrongLink
    :type job_id: ``int``
    :returns: True if job with job_id is finished; else False
    :rtype: ``bool`` or ``None``
    """
    job_status: StatusJob = get_job_status(job_id)
    if job_status is None:
        return None
    return job_status.is_finished()


def is_job_processing(job_id: int) -> typing.Optional[bool]:
    """Check the status of a tape

    :param job_id: id of a recall job in StrongLink
    :type job_id: ``int``
    :returns: True if job with job_id is currently being processed; else False
    :rtype: ``bool`` or ``None``
    """
    job_status: StatusJob = get_job_status(job_id)
    if job_status is None:
        return None
    return job_status.is_processing()


def is_job_queued(job_id: int) -> typing.Optional[bool]:
    """Check the status of a tape

    :param job_id: id of a recall job in StrongLink
    :type job_id: ``int``
    :returns: True if job with job_id is queued; else False
    :rtype: ``bool`` or ``None``
    """
    job_status: StatusJob = get_job_status(job_id)
    if job_status is None:
        return None
    return job_status.is_queued()


def job_queue() -> dict:
    """Prints status of the queue of tape read jobs

    :returns: information on the recall job queue (number of queued and processing jobs and some more information)
    :rtype: ``dict``
    """
    return json.loads(job_queue_raw(interpret="JSON"))



import inspect
import subprocess
from pathlib import Path
from typing import Union

from ..constants import PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS, SLK, SLK_HELPERS
from ..pyslk_exceptions import PySlkException
from ..utils import run_slk, which

__all__ = [
    "list_raw",
    "list_clone_file_raw",
    "list_clone_search_raw",
]


def list_raw(
    path_or_id: Union[str, int, Path],
    show_hidden: bool = False,
    numeric_ids: bool = False,
    recursive: bool = False,
    text: bool = False,
    print_bytes: bool = False,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """List results from search id or GNS path

    :param path_or_id: search id or gns path
    :type path_or_id: str or int or Path
    :param show_hidden: show '.' files, default: False (don't show these files)
    :type show_hidden: bool
    :param numeric_ids: show numeric values for user and group, default: False
        (show user and group names)
    :type numeric_ids: bool
    :param recursive: use the -R flag to list recursively, default: False
    :type recursive: bool
    :param text: print result to file 'slk_${USER}_list.txt', default: False
        (print to command line / print non-empty return string)
    :type text: bool
    :param print_bytes: use the -b to show sizes in bytes, default: False
    :type print_bytes: bool
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK) is None:
        raise PySlkException(f"pyslk: {SLK}: command not found")

    slk_call = [SLK, "list"]
    if show_hidden:
        slk_call.append("-a")
    if numeric_ids:
        slk_call.append("-n")
    if recursive:
        slk_call.append("-R")
    if text:
        slk_call.append("--text")
    if print_bytes:
        slk_call.append("-b")
    if isinstance(path_or_id, (str, Path)):
        slk_call.append(str(path_or_id))
    elif isinstance(path_or_id, int):
        slk_call.append(str(path_or_id))
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'path_or_id' need to be 'str', 'int' or Path-like but got "
            + f"'{type(path_or_id).__name__}'"
        )

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def list_clone_file_raw(
    resource_paths: Union[str, Path, list[str], list[Path], set[str], set[Path]],
    print_resource_ids: bool = False,
    print_timestamps_as_seconds_since_1970: bool = False,
    proceed_on_error: bool = False,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """List file information

    :param resource_paths: namespace or resource
    :type resource_paths: `str` or `Path`
    :param print_resource_ids: print resource ids instead of file paths
    :type print_resource_ids: `bool`
    :param print_timestamps_as_seconds_since_1970: print timestamps in seconds since 1970
    :type print_timestamps_as_seconds_since_1970: `bool`
    :param proceed_on_error: Proceed listing files even if an error arose
    :type proceed_on_error: `bool`
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]

    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    slk_call = [SLK_HELPERS, "list_clone_file"]

    if isinstance(resource_paths, str) and resource_paths != "":
        slk_call.append(resource_paths)
    elif isinstance(resource_paths, (list, set)) and len(resource_paths) > 0:
        if not all([isinstance(r, (str, Path)) for r in resource_paths]):
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: if argument 'resource_paths' is of type 'list' its items need to be"
                + "of type 'str' or Path-like but got type(s): "
                + f"'{', '.join([type(r).__name__ for r in resource_paths if not isinstance(r, (str, Path))])}'."
            )
        slk_call.extend([str(res) for res in resource_paths])
    elif isinstance(resource_paths, Path):
        slk_call.append(str(resource_paths))
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'resource_paths' has to be of type 'str', path-like or 'list'/'set' "
            + f"of these but is {type(resource_paths).__name__}"
        )

    if print_resource_ids:
        slk_call.append("--print-resource-ids")
    if print_timestamps_as_seconds_since_1970:
        slk_call.append("--print-timestamps-as-seconds-since-1970")
    if proceed_on_error:
        slk_call.append("--proceed-on-error")

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def list_clone_search_raw(
    search_id: Union[int, str],
    only_files: bool = False,
    only_namespaces: bool = False,
    start: Union[int, None] = None,
    count: Union[int, None] = None,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """List results from search id and important time stamps

    :param search_id: search id of search which results should be printed
    :type search_id: int or str
    :param only_files: print only files (like default for `slk list`)
    :type only_files: bool
    :param only_namespaces: print only namespaces
    :type only_namespaces: bool
    :param start: collect search results starting with the result 'start'; if set to 1 then collect from the beginning
    :type start: int
    :param count: Collect as many search results as defined by this parameter. If set to 0 then collect until end.
    :type count: int
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]

    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    slk_call = [SLK_HELPERS, "list_clone_search"]
    if only_files:
        slk_call.append("-f")
    if only_namespaces:
        slk_call.append("-d")
    if start is not None:
        slk_call.append("--start")
        slk_call.append(str(start))
    if count is not None:
        slk_call.append("--count")
        slk_call.append(str(count))
    if isinstance(search_id, (str, int)):
        slk_call.append(str(search_id))
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'search_id' has to be of type 'str' or 'int' but is "
            + f"{type(search_id).__name__}"
        )

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )



from .jobs import job_exists_raw, job_queue_raw, job_report_raw, job_status_raw
from .jobs_verify import result_verify_job_raw, submit_verify_job_raw
from .listing import list_clone_file_raw, list_clone_search_raw, list_raw
from .login import hostname_raw, session_raw
from .metadata import hsm2json_raw, json2hsm_raw, metadata_raw, tag_raw
from .resource import (
    checksum_raw,
    chmod_raw,
    delete_raw,
    exists_raw,
    group_raw,
    has_no_flag_partial_raw,
    mkdir_raw,
    move_raw,
    owner_raw,
    rename_raw,
    resource_mtime_raw,
    resource_path_raw,
    resource_permissions_raw,
    resource_tape_raw,
    resource_type_raw,
    size_raw,
)
from .search import (
    gen_file_query_raw,
    gen_search_query_raw,
    search_immediately_raw,
    search_incomplete_raw,
    search_raw,
    search_status_raw,
    search_successful_raw,
    searchid_exists_raw,
    tnsr_raw,
    total_number_search_results_raw,
)
from .storage import (
    group_files_by_tape_raw,
    is_on_tape_raw,
    iscached_raw,
    print_rcrs_raw,
    tape_barcode_raw,
    tape_exists_raw,
    tape_id_raw,
    tape_library_raw,
    tape_status_raw,
)
from .transfer import archive_raw, recall2_raw, recall_raw, retrieve2_raw, retrieve_raw
from .versioning import version_slk_helpers_raw, version_slk_raw

__all__ = [
    "archive_raw",
    "checksum_raw",
    "chmod_raw",
    "delete_raw",
    "exists_raw",
    "gen_file_query_raw",
    "gen_search_query_raw",
    "group_files_by_tape_raw",
    "group_raw",
    "has_no_flag_partial_raw",
    "hostname_raw",
    "hsm2json_raw",
    "is_on_tape_raw",
    "iscached_raw",
    "job_exists_raw",
    "job_queue_raw",
    "job_report_raw",
    "job_status_raw",
    "json2hsm_raw",
    "list_clone_file_raw",
    "list_clone_search_raw",
    "list_raw",
    "metadata_raw",
    "mkdir_raw",
    "move_raw",
    "owner_raw",
    "print_rcrs_raw",
    "recall_raw",
    "recall2_raw",
    "rename_raw",
    "resource_mtime_raw",
    "resource_path_raw",
    "resource_permissions_raw",
    "resource_tape_raw",
    "resource_type_raw",
    "result_verify_job_raw",
    "retrieve_raw",
    "retrieve2_raw",
    "search_immediately_raw",
    "search_incomplete_raw",
    "search_raw",
    "search_status_raw",
    "search_successful_raw",
    "searchid_exists_raw",
    "session_raw",
    "size_raw",
    "submit_verify_job_raw",
    "tag_raw",
    "tape_barcode_raw",
    "tape_exists_raw",
    "tape_id_raw",
    "tape_library_raw",
    "tape_status_raw",
    "tnsr_raw",
    "total_number_search_results_raw",
    "version_slk_helpers_raw",
    "version_slk_raw",
]



from __future__ import annotations

__all__ = ["StatusJob"]

import inspect
from typing import Union


class StatusJob:
    PAUSED, QUEUED, PROCESSING, COMPLETED, SUCCESSFUL, FAILED, ABORTED = range(-4, 3)
    STATI = {}

    # PROCESSING includes COMPLETING
    # ABORTED includes ABORTING

    def __init__(self, status_str: str):
        """
        Converts a string representing the status of a recall job into a status_job object

        Possible values for status_str are:
            PAUSED, PAUSING, QUEUED, PROCESSING, COMPLETED, COMPLETING, ABORTED, ABORTING

        :param status_str: status of a recall job
        :type status_str: str
        """
        if not isinstance(status_str, str):
            raise TypeError(
                f"pyslk.StatusJob.{inspect.stack()[0][3]}: wrong type provided; need 'str'; "
                + f"got '{type(status_str).__name__}'"
            )

        self.status: int

        self.STATI[self.PAUSED] = "PAUSED"
        self.STATI[self.QUEUED] = "QUEUED"
        self.STATI[self.PROCESSING] = "PROCESSING"
        self.STATI[self.COMPLETED] = "COMPLETED"
        self.STATI[self.SUCCESSFUL] = "SUCCESSFUL"
        self.STATI[self.FAILED] = "FAILED"
        self.STATI[self.ABORTED] = "ABORTED"

        if status_str == "SUCCESSFUL":
            self.status = self.SUCCESSFUL
        elif status_str == "FAILED":
            self.status = self.FAILED
        elif status_str == "COMPLETED":
            self.status = self.COMPLETED
        elif status_str in ["PROCESSING", "COMPLETING"]:
            self.status = self.PROCESSING
        elif status_str[0:6] == "QUEUED":
            self.status = self.QUEUED
        elif status_str in ["PAUSED", "PAUSING"]:
            self.status = self.PAUSED
        elif status_str in ["ABORTED", "ABORTING"]:
            self.status = self.ABORTED
        else:
            raise ValueError(
                f"pyslk.StatusJob.{inspect.stack()[0][3]}: provided status cannot be processed: {status_str}; "
                + "please contact support@dkrz.de"
            )

    def get_status(self) -> int:
        """
        return the status as integer

        Meaning of the output
        * -4: PAUSED / PAUSING
        * -3: QUEUED
        * -2: PROCESSING / COMPLETING
        * -1: COMPLETED
        * 0: SUCCESSFUL
        * 1: FAILED
        * 2: ABORTED / ABORTING

        :return: status as integer value (-4 to 2)
        :rtype: int
        """
        return self.status

    def get_status_name(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return self.STATI[self.status]

    def get_possible_stati(self) -> dict:
        return self.STATI

    def is_paused(self) -> bool:
        return self.status == self.PAUSED

    def is_queued(self) -> bool:
        return self.status == self.QUEUED

    def is_processing(self) -> bool:
        return self.status == self.PROCESSING

    def is_finished(self) -> bool:
        return self.status in [
            self.COMPLETED,
            self.SUCCESSFUL,
            self.FAILED,
            self.ABORTED,
        ]

    def is_successful(self) -> bool:
        return self.status == self.SUCCESSFUL

    def is_completed(self) -> bool:
        return self.status == self.COMPLETED

    def has_failed(self) -> bool:
        return self.status in [self.FAILED, self.ABORTED]

    def __eq__(self, other: Union[StatusJob, int]) -> bool:
        if isinstance(other, StatusJob):
            return self.status == other.get_status()
        if isinstance(other, int):
            return self.status == other
        raise TypeError(
            f"pyslk.StatusJob.{inspect.stack()[0][3]}: wrong type provided; need 'int' or 'StatusJob'; "
            + f"got '{type(other).__name__}'"
        )

    def __ne__(self, other: StatusJob) -> bool:
        if isinstance(other, StatusJob):
            return self.status != other.get_status()
        if isinstance(other, int):
            return self.status != other
        raise TypeError(
            f"pyslk.StatusJob.{inspect.stack()[0][3]}: wrong type provided; need 'int' or 'StatusJob'; "
            + f"got '{type(other).__name__}'"
        )



#! /usr/bin/env python
# -*- coding: utf-8 -*-
import logging
import os

"""
pyslk.constants provides constants for the pyslk package
"""

__all__ = [
    "MAX_RETRIES_RETRIEVAL_ON_TIMEOUT",
    "MAX_RETRIES_SUBMIT_JOBS_BASIC",
    "MAX_RETRIES_SUBMIT_JOBS_SAVE_MODE",
    "MIN_VERSION_SLK",
    "MIN_VERSION_SLK_HELPERS",
    "PYSLK_DEFAULT_LIST_CLONE_COLUMNS",
    "PYSLK_DEFAULT_LIST_COLUMNS",
    "PYSLK_DEFAULT_LIST_COLUMNS_3_3_21",
    "PYSLK_DEFAULT_LIST_COLUMN_TYPES",
    "PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS",
    "PYSLK_FILE_SIZE_UNITS",
    "PYSLK_LOGGER",
    "PYSLK_SEARCH_DELAY",
    "PYSLK_WILDCARDS",
    "SEARCH_QUERY_OPERATORS",
    "SLK",
    "SLK_HELPERS",
    "SLK_SYSTEM_CONFIG",
    "SLK_USER_CONFIG",
    "SLK_USER_LOG",
]

PYSLK_LOGGER = logging.getLogger("pyslk")
PYSLK_LOGGER.setLevel(logging.WARNING)
SLK_USER_CONFIG = os.path.expanduser("~/.slk/config.json")
SLK_SYSTEM_CONFIG = "/etc/stronglink.conf"
SLK_USER_LOG = os.path.expanduser("~/.slk/slk-cli.log")
MAX_RETRIES_SUBMIT_JOBS_BASIC = 5
MAX_RETRIES_SUBMIT_JOBS_SAVE_MODE = 5
MAX_RETRIES_RETRIEVAL_ON_TIMEOUT = 3

# wildcard will trigger a slk.list in core functions
# since slk can not handle wildcards in archive and retrieve calls
PYSLK_WILDCARDS = "*"

# if we want to retrieve several search ids from tape id groups,
# we should wait a little between grouping and retrieving because
# otherwise search ids will not be valid!
PYSLK_SEARCH_DELAY = 5

PYSLK_FILE_SIZE_UNITS = {"B": 1, "K": 10**3, "M": 10**6, "G": 10**9, "T": 10**12}
# PYSLK_FILE_SIZE_UNITS = {"B": 1, "K": 2**10, "M": 2**20, "G": 2**30, "T": 2**40}

# operators for gen_search_query
SEARCH_QUERY_OPERATORS: list[str] = ["=", "<", ">", "<=", ">="]

PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS = 3

PYSLK_DEFAULT_LIST_COLUMNS = [
    "permissions",
    "owner",
    "group",
    "filesize",
    "day",
    "month",
    "year",
    "time",
    "filename",
]
PYSLK_DEFAULT_LIST_COLUMNS_3_3_21 = [
    "permissions",
    "owner",
    "group",
    "filesize",
    "day",
    "month",
    "year",
    "filename",
]
PYSLK_DEFAULT_LIST_CLONE_COLUMNS = [
    "permissions",
    "owner",
    "group",
    "filesize",
    "timestamp_mtime",
    "timestamp_created",
    "timestamp_modified",
    "timestamp_storage",
    "tape_id",
    "filename",
]
PYSLK_DEFAULT_LIST_COLUMN_TYPES = {
    "owner": int,
    "group": int,
    "filesize": int,
    "tape_id": int,
}

SLK = "slk"
SLK_HELPERS = "slk_helpers"

MIN_VERSION_SLK = {
    "help": "3.3.21",
    "version": "3.3.0",
    "login": "3.0.0",
    "search": "3.3.0",
    "list": "3.3.56",
    "retrieve": "3.3.56",
    "recall": "3.1.0",
    "archive": "3.3.56",
    "tag": "3.3.56",
    "owner": "3.1.0",
    "group": "3.1.0",
    "chmod": "3.1.0",
    "move": "3.3.56",
    "rename": "3.0.0",
    "delete": "3.1.53",
}

MIN_VERSION_SLK_HELPERS = {
    "checksum": "1.0.0",
    "exists": "1.0.0",
    "gen_file_query": "1.8.4",
    "gen_search_query": "1.9.2",
    "gfbt": "1.13.3",
    "group_files_by_tape": "1.13.3",
    "has_no_flag_partial": "1.9.0",
    "help": "1.0.0",
    "hostname": "1.0.0",
    "hsm2json": "1.7.5",
    "is_admin_session": "1.9.9",
    "is_on_tape": "1.9.1",
    "iscached": "1.9.0",
    "job_exists": "1.7.0",
    "job_queue": "1.8.9",
    "job_report": "1.9.9",
    "job_status": "1.7.0",
    "json2hsm": "1.7.5",
    "list_clone_file": "1.13.0",
    "list_clone_search": "1.12.8",
    "list_search": "1.2.0",
    "metadata": "1.7.5",
    "mkdir": "1.0.0",
    "print_rcrs": "1.12.0",
    "recall": "1.13.0",
    "recall_needed": "1.13.0",
    "resource_id": "1.13.0",
    "resource_path": "1.8.0",
    "resource_permissions": "1.8.4",
    "resource_tape": "1.13.0",
    "resource_type": "1.8.4",
    "resourcepath": "1.0.0",
    "result_verify_job": "1.12.0",
    "retrieve": "1.13.0",
    "search_immediately": "1.9.9",
    "search_incomplete": "1.9.9",
    "search_limited": "1.1.0",
    "search_successful": "1.9.9",
    "search_status": "1.12.0",
    "searchid_exists": "1.8.8",
    "session": "1.0.0",
    "size": "1.10.0",
    "submit_verify_job": "1.12.3",
    "tape_barcode": "1.8.0",
    "tape_exists": "1.8.0",
    "tape_id": "1.8.0",
    "tape_library": "1.12.7",
    "tape_status": "1.8.0",
    "tnsr": "1.12.4",
    "total_number_search_results": "1.5.6",
    "version": "1.1.0",
}



import inspect
import subprocess
from pathlib import Path
from typing import Union

from ..constants import PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS, SLK, SLK_HELPERS
from ..pyslk_exceptions import PySlkException
from ..utils import run_slk, which

__all__ = [
    "resource_tape_raw",
    "exists_raw",
    "checksum_raw",
    "rename_raw",
    "move_raw",
    "delete_raw",
    "mkdir_raw",
    "size_raw",
    "resource_mtime_raw",
    "resource_path_raw",
    "resource_type_raw",
    "resource_permissions_raw",
    "has_no_flag_partial_raw",
    "owner_raw",
    "group_raw",
    "chmod_raw",
]


def resource_tape_raw(
    resource: str,
    json: bool = False,
    print_tape_barcode_only: bool = False,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """prints resource tape information

    :param resource: resource (full path)
    :type resource: `str`
    :param json: print json
    :type json: `bool`
    :param print_tape_barcode_only: print only barcode and no other information like tape id
    :type print_tape_barcode_only: `bool`
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    slk_call = [SLK_HELPERS, "resource_tape", resource]

    # --json
    if not isinstance(json, bool):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'json' has wrong type; need 'bool' but got "
            + f"'{type(json).__name__}'"
        )
    if json:
        slk_call.append("--json")

    # --print-tape-barcode-only
    if not isinstance(print_tape_barcode_only, bool):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'print_tape_barcode_only' has wrong type; need 'bool' but got "
            + f"'{type(print_tape_barcode_only).__name__}'"
        )
    if print_tape_barcode_only:
        slk_call.append("--print-tape-barcode-only")

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def exists_raw(
    gns_path: Union[str, Path], return_type: int = 0
) -> Union[str, int, subprocess.CompletedProcess]:
    """Check if resource exists

    :param gns_path: namespace or resource
    :type gns_path: str
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    if not isinstance(gns_path, (str, Path)):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'gns_path' needs to be 'str' or Path-like but got "
            + f"'{type(gns_path).__name__}'."
        )

    slk_call = [SLK_HELPERS, "exists", str(gns_path)]

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def checksum_raw(
    resource_path: Union[str, Path],
    checksum_type: Union[str, None] = None,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """Get a checksum of a resource

    :param resource_path: resource (full path)
    :type resource_path: str or Path
    :param checksum_type: checksum_type (possible values: None, "sha512",
                          "adler32"; None => print all)
    :type checksum_type: str
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    if not isinstance(resource_path, (str, Path)):
        raise TypeError(
            f'pyslk.{inspect.stack()[0][3]}: "resource_path" has to be of type "str" but is '
            + f"{type(resource_path).__name__}."
        )

    slk_call = [SLK_HELPERS, "checksum"]

    if checksum_type is not None:
        slk_call.append("-t")
        slk_call.append(checksum_type)
    slk_call.append(str(resource_path))

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def rename_raw(
    old_name: Union[str, Path], new_name: Union[str, Path], return_type: int = 0
) -> Union[str, int, subprocess.CompletedProcess]:
    """Rename a folder or file; moving is not possible

    :param old_name: folder or file name (full GNS path)
    :type old_name: str or Path
    :param new_name: new name (only name; no full GNS path)
    :type new_name: str or Path
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK) is None:
        raise PySlkException(f"pyslk: {SLK}: command not found")

    slk_call = [SLK, "rename"]
    if isinstance(old_name, (str, Path)):
        slk_call.append(str(old_name))
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'old_name' has to be of type 'str' or 'Path' but is "
            + f"{type(old_name).__name__}"
        )
    if isinstance(new_name, (str, Path)):
        slk_call.append(str(new_name))
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'new_name' has to be of type 'str' or 'Path' but is "
            + f"{type(new_name).__name__}"
        )

    if return_type == 0:
        return run_slk(slk_call, inspect.stack()[0][3])
    elif return_type == 1:
        return run_slk(slk_call, inspect.stack()[0][3], handle_output=False).returncode
    elif return_type == 2:
        return run_slk(slk_call, inspect.stack()[0][3], handle_output=False)
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def move_raw(
    src_path: Union[str, Path],
    dst_gns: Union[str, Path],
    interactive: bool,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """Move namespaces/files from one parent folder to another; renaming is
    not possible

    :param src_path: namespace or file (full GNS path)
    :type src_path: str or Path
    :param dst_gns: new parent namespace
    :type dst_gns: str or Path
    :param interactive: prompt user before overwrite; none-interactive: don't overwrite
    :type interactive: bool
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK) is None:
        raise PySlkException(f"pyslk: {SLK}: command not found")

    slk_call = [SLK, "delete"]
    if isinstance(src_path, (str, Path)):
        slk_call.append(str(src_path))
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'src_path' has to be of type 'str' or 'Path' but is "
            + f"{type(src_path).__name__}"
        )
    if isinstance(dst_gns, (str, Path)):
        slk_call.append(str(dst_gns))
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'dst_gns' has to be of type 'str' or 'Path' but is "
            + f"{type(dst_gns).__name__}"
        )
    if interactive:
        slk_call.append("-i")

    if return_type == 0:
        return run_slk(slk_call, inspect.stack()[0][3])
    elif return_type == 1:
        return run_slk(slk_call, inspect.stack()[0][3], handle_output=False).returncode
    elif return_type == 2:
        return run_slk(slk_call, inspect.stack()[0][3], handle_output=False)
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def delete_raw(
    gns_path: Union[Path, str, list], recursive: bool = False, return_type: int = 0
) -> Union[str, int, subprocess.CompletedProcess]:
    """Soft delete a namespace (optionally all child objects of a non-empty
            namespace) or a specific file

    :param gns_path: namespace or file (full GNS path); can be file list
    :type gns_path: str or list or Path
    :param recursive: use the -R flag to delete recursively, Default: False
    :type recursive: bool
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK) is None:
        raise PySlkException(f"pyslk: {SLK}: command not found")

    slk_call = [SLK, "delete"]
    if recursive:
        slk_call.append("-R")
    if isinstance(gns_path, (str, Path)):
        slk_call.append(str(gns_path))
    elif isinstance(gns_path, list):
        if not all([isinstance(r, (str, Path)) for r in gns_path]):
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: if argument 'gns_path' is of type 'list' its items need to be"
                + "of type 'str' or Path-like but got type(s): "
                + f"'{', '.join([type(r).__name__ for r in gns_path if not isinstance(r, (str, Path))])}'."
            )
        slk_call.extend(gns_path)
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'gns_path' needs to be 'str' or 'Path' but got "
            + f"{type(gns_path).__name__}."
        )

    if return_type == 0:
        return run_slk(slk_call, inspect.stack()[0][3])
    elif return_type == 1:
        return run_slk(slk_call, inspect.stack()[0][3], handle_output=False).returncode
    elif return_type == 2:
        return run_slk(slk_call, inspect.stack()[0][3], handle_output=False)
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def mkdir_raw(
    gns_path: Union[str, Path],
    recursive: bool = False,
    parents: bool = False,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """Create a directory

    :param gns_path: gns path to create
    :type gns_path: str or Path
    :param recursive: use the -R create folders recursively; throw error if folder already exists
    :type recursive: bool
    :param parents: create parent folders recursively; throw no error if folder already exists (like 'mkdir -p')
    :type parents: bool
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    if not isinstance(gns_path, (str, Path)):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'gns_path' needs to be 'str' or 'Path' but got "
            + f"{type(gns_path).__name__}."
        )

    slk_call = [SLK_HELPERS, "mkdir"]

    if recursive:
        slk_call.append("-R")
    if parents:
        slk_call.append("-p")
    slk_call.append(str(gns_path))

    if return_type == 0:
        return run_slk(slk_call, inspect.stack()[0][3])
    elif return_type == 1:
        return run_slk(slk_call, inspect.stack()[0][3], handle_output=False).returncode
    elif return_type == 2:
        return run_slk(slk_call, inspect.stack()[0][3], handle_output=False)
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def size_raw(
    resource_path: Union[str, Path, None] = None,
    resource_id: Union[str, int, None] = None,
    recursive: bool = False,
    pad_spaces_left: int = -1,
    verbose: bool = False,
    double_verbose: bool = False,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """Returns file size in byte

    :param resource_path: namespace or resource
    :type resource_path: `str` or `Path` or `None`
    :param resource_id: a resource id
    :type resource_id: `str` or `int`
    :param recursive: use the -R to calculate size recursively
    :type recursive: `bool`
    :param pad_spaces_left: pad spaces left
    :type pad_spaces_left: `int`
    :param verbose: single verbose mode, Default: False
    :type verbose: `bool`
    :param double_verbose: double verbose mode, Default: False
    :type double_verbose: `bool`
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: `int`
    :returns: stdout of the slk call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    if resource_id is not None and not isinstance(resource_id, (str, int)):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'resource_id' has to be of type 'str' or 'int' but is "
            + f"{type(resource_id).__name__}"
        )

    if resource_path is not None and not isinstance(resource_path, (str, Path)):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'resource_path' has to be of type 'str' or path-like but is "
            + f"{type(resource_path).__name__}"
        )

    slk_call = [SLK_HELPERS, "size"]

    if resource_path is not None:
        slk_call.append(str(resource_path))
    if resource_id is not None:
        slk_call.append("--resource-id")
        slk_call.append(str(resource_id))
    # -R / --recursive
    if recursive:
        slk_call.append("-R")
    # --pad-spaces-left
    if not isinstance(pad_spaces_left, int):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: 'pad_spaces_left' has to be of type 'int' but is "
            + f"{type(pad_spaces_left).__name__}"
        )
    if pad_spaces_left != -1:
        slk_call.append("--pad-spaces-left")
        slk_call.append(str(pad_spaces_left))
    # -vv (double verbose)
    if double_verbose:
        slk_call.append("-vv")
    elif verbose:
        # -v (verbose)
        slk_call.append("-v")

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def resource_path_raw(
    resource_id: Union[str, int], return_type: int = 0
) -> Union[str, int, subprocess.CompletedProcess]:
    """Get path for a resource id

    :param resource_id: a resource_id
    :type resource_id: str or int
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    if not isinstance(resource_id, (str, int)):
        raise TypeError(
            f'pyslk.{inspect.stack()[0][3]}: "resource_id" has to be of type "str" or "int" but is '
            + f"{type(resource_id).__name__}"
        )

    slk_call = [SLK_HELPERS, "resource_path", str(resource_id)]

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def resource_type_raw(
    resource_path: Union[str, Path, None] = None,
    resource_id: Union[str, int, None] = None,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """Get type of resource provided via resource id or a resource path

    :param resource_path: resource path (either resource_id or resource_path have to be provided; not both)
    :type resource_path: str or path-like or None
    :param resource_id: a resource id
    :type resource_id: str or int
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    if resource_id is not None and not isinstance(resource_id, (str, int)):
        raise TypeError(
            f'pyslk.{inspect.stack()[0][3]}: "resource_id" has to be of type "str" or "int" but is '
            + f"{type(resource_id).__name__}"
        )

    if resource_path is not None and not isinstance(resource_path, (str, Path)):
        raise TypeError(
            f'pyslk.{inspect.stack()[0][3]}: "resource_path" has to be of type "str" or path-like but is '
            + f"{type(resource_path).__name__}"
        )

    slk_call = [SLK_HELPERS, "resource_type"]

    if resource_path is not None:
        slk_call.append(str(resource_path))
    if resource_id is not None:
        slk_call.append("--resource-id")
        slk_call.append(str(resource_id))

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def resource_mtime_raw(
    resource_path: Union[str, Path, None] = None,
    resource_id: Union[str, int, None] = None,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """Get mtime of resource provided via resource id or a resource path

    :param resource_path: resource path (either resource_id or resource_path have to be provided; not both)
    :type resource_path: str or path-like or None
    :param resource_id: a resource id
    :type resource_id: str or int
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    if resource_id is not None and not isinstance(resource_id, (str, int)):
        raise TypeError(
            f'pyslk.{inspect.stack()[0][3]}: "resource_id" has to be of type "str" or "int" but is '
            + f"{type(resource_id).__name__}"
        )

    if resource_path is not None and not isinstance(resource_path, (str, Path)):
        raise TypeError(
            f'pyslk.{inspect.stack()[0][3]}: "resource_path" has to be of type "str" or path-like but is '
            + f"{type(resource_path).__name__}"
        )

    slk_call = [SLK_HELPERS, "resource_mtime"]

    if resource_path is not None:
        slk_call.append(str(resource_path))
    if resource_id is not None:
        slk_call.append("--resource-id")
        slk_call.append(str(resource_id))

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def resource_permissions_raw(
    resource_path: Union[str, Path, None] = None,
    resource_id: Union[str, int, None] = None,
    as_octal_number: bool = False,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """Get permissions of a resource provided via resource id or a resource path

    :param resource_path: resource path (either resource_id or resource_path have to be provided; not both)
    :type resource_path: str or path-like
    :param resource_id: a resource id
    :type resource_id: str or int
    :param as_octal_number: Do not return the permissions as combination of x/w/r/- but as three digit octal number
    :type as_octal_number: bool
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    if resource_id is not None and not isinstance(resource_id, (str, int)):
        raise TypeError(
            f'pyslk.{inspect.stack()[0][3]}: "resource_id" has to be of type "str" or "int" but is '
            + f"{type(resource_id).__name__}"
        )

    if resource_path is not None and not isinstance(resource_path, (str, Path)):
        raise TypeError(
            f'pyslk.{inspect.stack()[0][3]}: "resource_path" has to be of type "str" or path-like but is '
            + f"{type(resource_path).__name__}"
        )

    slk_call = [SLK_HELPERS, "resource_permissions"]

    if resource_path is not None:
        slk_call.append(str(resource_path))
    if resource_id is not None:
        slk_call.append("--resource-id")
        slk_call.append(str(resource_id))
    if as_octal_number is True:
        slk_call.append("--as-octal-number")

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def has_no_flag_partial_raw(
    resource_path: Union[str, Path, list, None] = None,
    resource_id: Union[str, int, None] = None,
    search_id: Union[str, int, None] = None,
    recursive: bool = False,
    verbose: bool = False,
    double_verbose: bool = False,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """Get info whether file is flagged as partial file

    :param resource_path: resource (full path)
    :type resource_path: str or Path or list or None
    :param resource_id: a resource id
    :type resource_id: str or int
    :param search_id: id of a search
    :type search_id: int or str
    :param recursive: go through all files in resource_path recursively
    :type recursive: bool
    :param verbose: single verbose mode, Default: False
    :type verbose: bool
    :param double_verbose: double verbose mode, Default: False
    :type double_verbose: bool
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK_HELPERS) is None:
        raise PySlkException(f"pyslk: {SLK_HELPERS}: command not found")

    if resource_id is not None and not isinstance(resource_id, (str, int)):
        raise TypeError(
            f'pyslk.{inspect.stack()[0][3]}: "resource_id" has to be of type "str" or "int" but is '
            + f"{type(resource_id).__name__}"
        )

    if search_id is not None and not isinstance(search_id, (str, int)):
        raise TypeError(
            f'pyslk.{inspect.stack()[0][3]}: "search_id" has to be of type "str" or "int" but is '
            + f"{type(resource_id).__name__}"
        )

    slk_call = [SLK_HELPERS, "has_no_flag_partial"]

    if resource_path is not None:
        if isinstance(resource_path, (str, Path)):
            slk_call.append(str(resource_path))
        elif isinstance(resource_path, list):
            if not all([isinstance(r, (str, Path)) for r in resource_path]):
                raise TypeError(
                    f"pyslk.{inspect.stack()[0][3]}: if argument 'resource_path' is of type 'list' its items need to be"
                    + "of type 'str' or Path-like but got type(s): "
                    + f"'{', '.join([type(r).__name__ for r in resource_path if not isinstance(r, (str, Path))])}'."
                )
            slk_call.extend(resource_path)
        else:
            raise TypeError(
                f'pyslk.{inspect.stack()[0][3]}: "resource_path" has to be of type "str", path-like or "list" but is '
                + f"{type(resource_path).__name__}"
            )
    if resource_id is not None:
        slk_call.append("--resource-id")
        slk_call.append(str(resource_id))
    if search_id is not None:
        slk_call.append("--search-id")
        slk_call.append(str(search_id))
    if recursive:
        slk_call.append("-R")
    if double_verbose:
        slk_call.append("-vv")
    elif verbose:
        slk_call.append("-v")

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def owner_raw(
    gns_path: Union[str, Path],
    owner: Union[str, int],
    recursive: bool = False,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """Change the owner of a resource or namespace

    :param gns_path: namespace or file (full GNS path)
    :type gns_path: str or Path
    :param owner: new owner of a file (username or uid)
    :type owner: str or int
    :param recursive: use the -R flag to delete recursively, Default: False
    :type recursive: bool
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK) is None:
        raise PySlkException(f"pyslk: {SLK}: command not found")

    # basic call
    slk_call = [SLK, "owner"]

    # check input types
    if isinstance(owner, (str, int)):
        slk_call.append(str(owner))
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'owner' needs to be 'str' or 'int' but got "
            + f"'{type(owner).__name__}'"
        )
    # add further arguments
    if recursive:
        slk_call.append("-R")
    if isinstance(gns_path, (str, Path)):
        slk_call.append(str(gns_path))
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'gns_path' needs to be 'str' or path-like but got "
            + f"'{type(gns_path).__name__}'"
        )

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def group_raw(
    gns_path: Union[str, Path],
    group: Union[str, int],
    recursive: bool = False,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """Change the group of a resource or namespace

    :param gns_path: namespace or file (full GNS path)
    :type gns_path: str or Path
    :param group: new group of a file (group name or gid)
    :type group: str or int
    :param recursive: use the -R flag to delete recursively, Default: False
    :type recursive: bool
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK) is None:
        raise PySlkException(f"pyslk: {SLK}: command not found")

    # basic call
    slk_call = [SLK, "group"]

    # check input types
    if isinstance(group, (str, int)):
        slk_call.append(str(group))
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'group' needs to be 'str' or 'int' but got "
            + f"'{type(group).__name__}'"
        )
    # add further arguments
    if recursive:
        slk_call.append("-R")
    if isinstance(gns_path, (str, Path)):
        slk_call.append(str(gns_path))
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'gns_path' needs to be 'str' or path-like but got "
            + f"'{type(gns_path).__name__}'"
        )

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )


def chmod_raw(
    gns_path: Union[str, list, Path],
    mode: Union[str, int],
    recursive: bool = False,
    return_type: int = 0,
) -> Union[str, int, subprocess.CompletedProcess]:
    """Change the access mode of a resource or namespace

    :param gns_path: namespace or file (full GNS path); can be file list
    :type gns_path: str or list or Path
    :param mode: new mode/permissions of a file (as known from bash's chmod)
    :type mode: str or int
    :param recursive: use the -R flag to delete recursively, Default: False
    :type recursive: bool
    :param return_type: select between 0 (== str output), 1 (== exit code), 2 (subprocess output)
    :type return_type: int
    :returns: stdout of the slk_helpers call
    :rtype: Union[str, int, subprocess.CompletedProcess]
    """
    if which(SLK) is None:
        raise PySlkException(f"pyslk: {SLK}: command not found")

    # basic call
    slk_call = [SLK, "chmod"]

    # check input types
    if isinstance(mode, (str, int)):
        slk_call.append(str(mode))
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'mode' needs to be 'str' or 'int' but got "
            + f"'{type(mode).__name__}'"
        )
    # add further arguments
    if recursive:
        slk_call.append("-R")
    if isinstance(gns_path, (str, Path)):
        slk_call.append(str(gns_path))
    elif isinstance(gns_path, list):
        if not all([isinstance(r, (str, Path)) for r in gns_path]):
            raise TypeError(
                f"pyslk.{inspect.stack()[0][3]}: if argument 'gns_path' is of type 'list' its items need to be"
                + "of type 'str' or Path-like but got type(s): "
                + f"'{', '.join([type(r).__name__ for r in gns_path if not isinstance(r, (str, Path))])}'."
            )
        slk_call.extend(gns_path)
    else:
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'gns_path' needs to be 'str', 'list' or path-like but got "
            + f"'{type(gns_path).__name__}'"
        )

    if return_type == 0:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
        )
    elif return_type == 1:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        ).returncode
    elif return_type == 2:
        return run_slk(
            slk_call,
            inspect.stack()[0][3],
            retries_on_timeout=PYSLK_DEFAULT_NUMBER_RETRIES_NONMOD_CMDS,
            handle_output=False,
        )
    else:
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_type' needs to be 0, 1 or 2."
        )



from .file_group import FileGroup
from .group_collection import GroupCollection
from .listing import list_clone_file, list_clone_search, ls
from .login import expiration_date, hostname, session, valid_session, valid_token
from .resource_path import get_resource_path
from .searching import (
    get_search_status,
    is_search_incomplete,
    is_search_successful,
    search,
    search_immediately,
    searchid_exists,
    total_number_search_results,
)
from .stati import StatusJob
from .versioning import cli_versions, version_slk, version_slk_helpers

__all__ = [
    "FileGroup",
    "GroupCollection",
    "StatusJob",
    "cli_versions",
    "expiration_date",
    "get_resource_path",
    "get_search_status",
    "hostname",
    "is_search_incomplete",
    "is_search_successful",
    "list_clone_file",
    "list_clone_search",
    "ls",
    "search",
    "search_immediately",
    "searchid_exists",
    "session",
    "total_number_search_results",
    "valid_session",
    "valid_token",
    "version_slk",
    "version_slk_helpers",
]



"""Configuration module of slk."""

from __future__ import annotations

import os
import re
import shutil
import warnings
from pathlib import Path
from subprocess import PIPE, CalledProcessError, run
from typing import Any, Literal, Optional

from .constants import PYSLK_LOGGER, SLK, SLK_HELPERS

__all__ = ["get", "set", "module_load"]


class set:
    """Set pyslk configuration to intract with the slk backend.

    This class can be used to set specifics on the the backend version
    of slk, for example the path the th slk binary or the slk version.

    Parameters
    ----------
    **kwargs: str | Path
        key-value pairs of the config values to set.

    Examples
    --------
    Set the specific slk version that is loaded on startup:

    ::

        from pyslk import pyslk
        pyslk.config.set(slk_version="3.3.76")


    It would be also possible to set the path to the slk binary. By default
    the the set class checks the path to the binary of the slk command. To
    set a custom path you can use the ``slk`` key. This time we are using
    the context manager to temporarely override the default path.

    ::

        from pyslk import pyslk
        with pyslk.config.set(
            slk="/sw/spack-levante/slk-3.3.73-nl4md3/bin/slk",
            slk_helpers="/sw/spack-levante/slk-3.3.73-nl4md3/bin/slk_helpers"):
                print(pyslk.config.get("__slk_version__"))
                print(pyslk.config.get("__slk_helpers_version__"))
        print(pyslk.config.get("__slk_version__"))
        print(pyslk.config.get("__slk_helpers_version__"))

        #
        3.3.73
        1.5.7
        3.3.21
        1.2.4
    """

    config: dict[str, Optional[str]] = {
        SLK: SLK,
        SLK_HELPERS: SLK_HELPERS,
        "module": SLK,
        "module_helpers": SLK_HELPERS,
    }
    """Default slk/slk_helpers commands, stored in the config cache. These
    can either be commands (default) or abs paths to the binaries."""
    _record: list[tuple[Literal["insert", "replace"], str, Optional[str]]]

    def __init__(self, **kwargs: str | Path):
        self._record = []
        for key, value in kwargs.items():
            self._assign(key.lower(), str(value), self.config)

    def __enter__(self) -> set:
        return self

    def __exit__(self, *args: Any) -> None:
        for op, key, value in reversed(self._record):
            if op == "replace":
                self.config[key] = value
            else:  # insert
                self.config.pop(key, None)

    def _assign(
        self,
        key: str,
        value: str,
        cfg: dict[str, Optional[str]],
        record: bool = True,
    ):
        if record:
            if key in cfg:
                self._record.append(("replace", key, cfg[key]))
            else:
                self._record.append(("insert", key, None))
        cfg[key] = value

    def get_path(self, key: str) -> str | None:
        """Get the path to the binary of slk"""
        value = self.config.get(key, key)
        if value is None:
            return None
        if Path(value).expanduser().is_absolute():
            return str(Path(self.config[key]).expanduser())
        slk_path = shutil.which(value)
        if slk_path:
            return slk_path
        return None

    def get_version(self, command) -> str:
        """Get the version of a command"""
        slk_path = self.get_path(SLK)
        env = module_load()
        if slk_path:
            env["PATH"] = f"{Path(slk_path).parent}" + os.pathsep + env["PATH"]
        try:
            return re.search(
                r"\d+\.\d+\.\d+",
                (
                    run(
                        [command, "version"],
                        env=env,
                        stdout=PIPE,
                        stderr=PIPE,
                    ).stdout.decode("utf-8")
                ),
            ).group()
        except (CalledProcessError, FileNotFoundError, AttributeError):
            warnings.warn(f"failed to dertermine {command} version")
            return "0.0.0"


def module_load(
    module_path: str = "/usr/share/Modules/libexec/modulecmd.tcl",
    _env: dict[str, dict[str, str]] = {},
) -> dict[str, str]:
    """Use the module load command to activate the slk environment."""
    slk_version = get("slk_version", "")
    module_name = get("module", SLK)
    if slk_version:
        module_name = f"{module_name}/{slk_version}"
    # Check if slk is already present in the current path, of so
    # let's assume the user want's to use this version then
    if _config.get_path(SLK) is not None:
        PYSLK_LOGGER.debug("slk already in users path, skipping module load")
        return os.environ.copy()
    if _env.get(module_name) is None:
        PYSLK_LOGGER.debug("Loading slk module %s", module_name)
        _env[module_name] = {}
        try:
            res = (
                run(
                    [module_path, "python", "load", module_name],
                    check=True,
                    stdout=PIPE,
                    stderr=PIPE,
                )
                .stdout.decode()
                .split("\n")
            )
        except (FileNotFoundError, CalledProcessError) as error:
            warnings.warn(f"Could not load {module_name}: {error}")
            return {}
        for line in res:
            if "=" in line:
                key, _, value = line.partition("=")
                new_key = (re.findall(r"\['([^']*)", key.strip()) or [""])[0]
                if new_key.strip():
                    _env[module_name][new_key.strip()] = value.strip().replace("'", "")
    else:
        PYSLK_LOGGER.debug("module %s has already been loaded", module_name)
    return _env[module_name]


_config = set()


def get(key: str, default: Optional[str] = None) -> Optional[float]:
    """Get global slk config.

    Parameters
    ----------
    key: str
        config parameter to query
    default: str, default: None
        default return value
    """
    if key == "__slk_version__":
        return _config.get_version(SLK)
    if key == "__slk_helpers_version__":
        return _config.get_version(SLK_HELPERS)
    return _config.config.get(key.lower(), default)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""\
pyslk.utils contains utilities needed by pyslk.pyslk.
"""

import calendar
import glob
import inspect
import json
import math
import os
import re
import shutil
import subprocess
import time
import warnings
from datetime import datetime
from grp import getgrgid
from pathlib import Path
from pwd import getpwuid
from stat import filemode
from typing import Optional, Union

import pandas as pd
import psutil
from dateutil.parser import parse

from .config import module_load
from .constants import (
    PYSLK_DEFAULT_LIST_CLONE_COLUMNS,
    PYSLK_DEFAULT_LIST_COLUMNS,
    PYSLK_FILE_SIZE_UNITS,
    PYSLK_LOGGER,
    PYSLK_WILDCARDS,
    SLK_USER_CONFIG,
    SLK_USER_LOG,
)
from .pyslk_exceptions import (
    PySlkBadProcessError,
    PySlkException,
    PySlkNoValidLoginTokenError,
    check_for_errors,
)

__all__ = [
    "construct_dst_from_src",
    "convert_expiration_date",
    "login_valid",
    "run_slk",
    "access_local",
    "is_search_id",
    "which",
    "get_recall_job_id",
    "get_slk_pid",
    "FakeProc",
    "_convert_size",
    "_parse_list",
    "_parse_list_to_rows",
    "_parse_size",
    "_rows_to_dataframe",
]


class FakeProc:
    i = 0

    def __init__(self):
        pass

    def next_i(self):
        if self.i < 6:
            self.i = self.i + 1
        else:
            self.i = 1
        return self.i


def which(command: str) -> str:
    """Get the path to a given command."""
    path = module_load().get("PATH")
    return shutil.which(command, path=path) or ""


def run_slk(
    slk_call: Union[list, str],
    fun: str,
    env: Optional[str] = None,
    retries_on_timeout: int = 0,
    handle_output: bool = True,
    wait_until_finished: bool = True,
) -> Union[str, subprocess.CompletedProcess, subprocess.Popen]:
    """Runs a provided slk_call via subprocess.run and returns stdout

    The argument 'slk_call' is used as argument in a call of
    'subprocess.run(...)'. If slk_call is a string then 'shell=True' is set in
    the run(...)-call. Otherwise, the argument 'shell' is omitted.

    If the command provided via 'slk_call' does not exist (e.g. 'slk' not
    available on this system) or if 'slk' returns an exit code '!= 0' then a
    PySlkException is raised.

    :param slk_call: slk call as input for subprocess.run
    :type slk_call: str or list of str
    :param fun: name of the calling pyslk function; printed when exception is
                raised
    :type fun: str
    :param env: environment string
    :type env: str
    :param retries_on_timeout: run slk_helpers calls repeatedly 'n' times when exit code 3 (timeout) is returned
    :type retries_on_timeout: int
    :param handle_output: whether to handle exceptions and parse outputs or just return outputs.
    :type handle_output: bool
    :param wait_until_finished: let this function run as long as the slk command is running
    :type wait_until_finished: bool
    :returns: stdout of the slk call
    :rtype: str
    """
    PYSLK_LOGGER.debug(f"slk call: {slk_call}")
    # set / get / update environment
    env = env or module_load()
    os_env = os.environ.copy()
    os_env.update(env)
    # convert slk_call to string list if it is a list
    if isinstance(slk_call, list):
        tmp_slk_call: list = [str(c) for c in slk_call]
        slk_call = tmp_slk_call
    # check if 'retries_on_timeout' is set to invalid value
    if retries_on_timeout < 0:
        raise ValueError("argument 'retries_on_timeout' has to be a positive integer")
    if retries_on_timeout > 10:
        warnings.warn(
            f"argument 'retries_on_timeout' is set to {retries_on_timeout} which might cause long waiting "
            + "times if the StrongLink system is actually not responding."
        )
    # count number of iterations
    i: int = 0
    # set 'True' if timeout error arises to remain in 'while' loop; needs to be initially 'True'; otherwise we do
    # not enter the while loop
    timeout_error: bool = True
    try:
        output = None
        while i <= retries_on_timeout and timeout_error:
            if isinstance(slk_call, list):
                if wait_until_finished:
                    output = subprocess.run(
                        slk_call,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        env=os_env,
                    )
                else:
                    output = subprocess.Popen(
                        slk_call,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        env=os_env,
                        text=True,
                    )
                    time.sleep(2)
            elif isinstance(slk_call, str):
                if wait_until_finished:
                    output = subprocess.run(
                        slk_call,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        shell=True,
                        env=os_env,
                    )
                else:
                    output = subprocess.Popen(
                        slk_call,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        shell=True,
                        env=os_env,
                        text=True,
                    )
                    time.sleep(2)
            else:
                raise TypeError(
                    f"pyslk.{fun}: 'slk_call' has to be of type 'list' or 'str' but is {type(slk_call).__name__}"
                )
            # if the return code is unequals 3 we leave the while loop by setting 'timeout_error' to False
            if output.returncode != 3:
                timeout_error = False
    except FileNotFoundError as e:
        raise PySlkException(
            f"pyslk.{fun}: {str(e).split(':')[-1][2:-1]}: command not found"
        )
    if handle_output and wait_until_finished:
        check_for_errors(output, fun)

        # get return value and ...
        tmp_return = output.stdout.decode("utf-8")

        # ... pimp a bit if necessary
        if len(tmp_return) > 0:
            if tmp_return[-1] == "\n":
                tmp_return = tmp_return[:-1]

        # return the return value
        return tmp_return
    return output


def get_slk_pid(process: Union[subprocess.Popen, FakeProc]) -> int:
    """
    Return the pid of a slk call

    :param process: a process as instance of subprocess.Popen in which slk is running as child process
    :type process: subprocess.Popen
    :return: pid of a slk call within the provided instance of subprocess.Popen
    :rtype:
    """
    if isinstance(process, FakeProc):
        return process.next_i()

    # child pid
    cpid: int = -1
    if process.poll() is None:
        try:
            parent = psutil.Process(process.pid)
        except psutil.NoSuchProcess:
            tmp_args = process.args
            if not isinstance(tmp_args, str) and tmp_args is not None:
                tmp_args = " ".join(tmp_args)
            raise PySlkBadProcessError(
                f"pyslk.{inspect.stack()[0][3]}: process with pid {str(process.pid)} and name {tmp_args} does not exist"
            )
        children = parent.children(recursive=True)
        for process in children:
            # if multiple processes with the name "java" exist, we take the last one
            if process.name()[0:4] == "java":
                cpid = process.pid
        if cpid == -1:
            tmp_args = process.args
            if not isinstance(tmp_args, str) and tmp_args is not None:
                tmp_args = " ".join(tmp_args)
            raise PySlkBadProcessError(
                f"pyslk.{inspect.stack()[0][3]}: process with pid {str(process.pid)} and name {tmp_args} has no child "
                + "process which looks like a running instance of slk/slk_helpers."
            )
    else:
        tmp_args = process.args
        if not isinstance(tmp_args, str) and tmp_args is not None:
            tmp_args = " ".join(tmp_args)
        raise PySlkBadProcessError(
            f"pyslk.{inspect.stack()[0][3]}: process with pid {str(process.pid)} and name {tmp_args} has already ended."
            + " Cannot get child processes."
        )
    # return child pid
    return cpid


def get_recall_job_id(
    pid: Union[int, str],
    return_last_match: bool = True,
    slk_log_file: Union[str, Path] = SLK_USER_LOG,
) -> Union[int, None]:
    """

    :param pid: pid of a slk call
    :type pid: int or str
    :param return_last_match: if multiple slk runs hat the same PID and returned a job id, take the last job id;
        else, take the first one
    :type return_last_match: bool
    :param slk_log_file: path of the slk-cli log file from which this command should read
    :type slk_log_file: str or path-like object
    :return: a valid job id; if no job id was printed to the slk-cli log, print None
    :rtype: int or None
    """
    text = str(pid) + r"\s*INFO\s*Created copy job with id:\s*'([0-9]*)' for"
    jid: int = -1
    with open(slk_log_file) as file:
        for line in file:
            if re.search(text, line):
                # if the pid existed  multiple times, we will get the most recent output
                jid = int(re.search(text, line).group(1))
                if not return_last_match:
                    return jid
    if jid == -1:
        return None
    else:
        return jid


def _parse_list_to_rows(
    output,
    path_or_id: Union[str, int, Path, None] = None,
    full_path: bool = True,
    is_resource_id: bool = False,
) -> list:
    """Parsed the output from a list command to a list of rows.

    The parsing is very much static and fixed. The parsing is done line by line
    to also catch recursive outputs. Filenames are concatenated with the path.
    If the string output originates from a recursive list or search,
    full_path should be True to distinguish between files of the same name.


    :param is_resource_id: the last column is a resource id - no resource path
    :type is_resource_id: bool
    TODO: add variable description
    """  # .format(slk_version_used_for_parser_syntax)
    rows = []
    concat_path = ""

    if full_path and isinstance(path_or_id, (str, Path)):
        concat_path = str(path_or_id)
        base = os.path.basename(concat_path)
        dirname = os.path.dirname(concat_path)
        if PYSLK_WILDCARDS in base:
            # only use dirname to concat in this case
            concat_path = os.path.dirname(concat_path)
        if PYSLK_WILDCARDS in dirname:
            # output slk list is ambiguous in this case, no concat_path
            concat_path = ""

    if path_or_id is None:
        warnings.warn(
            "When 'path_or_id' is not set (which is the case), 'full_path' will be ignored."
        )

    for line in output:
        # if this is from a recursive list
        if line.startswith("/") and line.endswith(":") and full_path is True:
            # we keep the path for later concatenation to filename
            concat_path = line[:-1]
        split = [entry for entry in line.split(" ") if entry]
        if len(split) == len(PYSLK_DEFAULT_LIST_COLUMNS):
            if not is_resource_id:
                # if we list a search result, then the path is printed automatically
                # thus, if we have a search_id, then we need to remove the path when full_path is False
                if isinstance(path_or_id, int) and not full_path:
                    split[-1] = os.path.basename(split[-1])
                # if path is a path and not a search id
                if isinstance(path_or_id, (str, Path)):
                    if len(output) == 3 and os.path.basename(concat_path) == split[-1]:
                        # in this case, we list only one item. there is no way to say
                        # if this is a directory containing only one file or simply a single file listed.
                        # we assume it's as single file listed. This only doesn't work if we actually
                        # have a single file in a directory named like the file. we assume this does not happen...
                        split[-1] = concat_path
                    else:
                        split[-1] = os.path.join(concat_path, split[-1])
            rows.append(split)
        if len(split) == len(PYSLK_DEFAULT_LIST_CLONE_COLUMNS):
            if not is_resource_id:
                # we need to remove the path when full_path is False
                if path_or_id is not None and not full_path:
                    split[-1] = os.path.basename(split[-1])
            rows.append(split)

    return rows


def _parse_list_to_rows_3_3_21(output, path_or_id="", full_path=True):
    """Parsed the output from a list command to a list of rows.

    The parsing is very much static and fixed. The parsing is done line by line
    to also catch recursive outputs. Filenames are concatenated with the path.
    If the string output originates from a recursive list or search,
    full_path should be True to distinguish between files of the same name.

    """  # .format(slk_version_used_for_parser_syntax)
    rows = []
    concat_path = ""
    if full_path is True:
        concat_path = path_or_id

    for line in output:
        # if this is from a recursive list
        if line.startswith("/") and line.endswith(":") and full_path is True:
            # we keep the path for later concatenation to filename
            concat_path = line[:-1]
        split = [entry for entry in line.split(" ") if entry]
        if len(split) == 7:
            split.insert(3, "")
        if len(split) == 8:
            # if path is a path and not a search id...
            # if type(path) is str:
            split[-1] = os.path.join(concat_path, split[-1])
            # _parse_date(split)
            rows.append(split)
            # rows.append(split)

    return rows


def _parse_size(size_str):
    """parse slk filesize format into integer bytes.

    Assumes slk format, e.g., 22.6G, 11M, 343K etc..

    If parsing fails, return NaN.

    """
    # import numpy as np
    units = PYSLK_FILE_SIZE_UNITS
    if size_str:
        if re.match("^[0-9]+$", size_str) is not None:
            return float(size_str)
        else:
            try:
                number, unit = size_str[0:-1], size_str[-1]
                return float(number) * units[unit]
            except Exception as e:
                warnings.warn(
                    f"pyslk.{inspect.stack()[0][3]}: parsing size failed with: {repr(e)}"
                )
                return math.nan
            # removed "int"-casting of the results
    warnings.warn(
        f"pyslk.{inspect.stack()[0][3]}: file size cannot be "
        f"converted: {size_str}; return NaN instead"
    )
    return math.nan


def _convert_size(size_byte: int, unit: str = "B") -> dict[str, Union[str, float, int]]:
    """
    Converts the size of a file from byte to chosen unit

    :param size_byte:
    :type size_byte:
    :param unit:
    :type unit:
    :return:
    :rtype: int
    """
    if unit == "h":
        if size_byte == 0:
            return {"value": 0, "unit": "B"}
        for best_unit in PYSLK_FILE_SIZE_UNITS.keys():
            best_size = size_byte / PYSLK_FILE_SIZE_UNITS[best_unit]
            if (best_size >= 1.0) & (best_size < 1000.0):
                return {"value": best_size, "unit": best_unit}
        warnings.warn(
            f"pyslk.{inspect.stack()[0][3]}: could not find "
            + f"out human-readable value using available units for value {str(size_byte)}"
        )
        return {"value": size_byte, "unit": "B"}
    if unit not in PYSLK_FILE_SIZE_UNITS.keys():
        raise ValueError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'return_format' has to be a string of value "
            + f"{', '.join(PYSLK_FILE_SIZE_UNITS.keys())} or 'h' (human-readable)."
        )
    # Return result in desired format
    new_size = size_byte / PYSLK_FILE_SIZE_UNITS[unit]
    return {"value": new_size, "unit": unit}


def _filename_list_2_search_query(directory, filename_list):
    """generates a RQL search query to find files with names "filename_list"
    located in "directory"

    We get a directory and a list of filenames and create a RQL search
    query.

    :param directory: a directory
    :type directory: str
    :param filename_list: list of filenames
    :type filename_list: list of str
    :returns: search query to find the files
    :rtype: str
    """

    if not isinstance(filename_list, list):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'filename_list' has to be a list of strings but has type "
            + f"{type(filename_list).__name__}"
        )

    if not isinstance(directory, str):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'directory' has to be a string but has type "
            + f"{type(directory).__name__}"
        )

    if len(filename_list) == 0:
        return '{{"path": {{"$gte": "{0}", "$max_depth": 1}}}}'.format(directory)

    tmp_resource_names: str = '"}, {"resources.name": "'.join(filename_list)
    return (
        f'{{"$and": [{{"path": {{"$gte": "{directory}", "$max_depth": 1}}}}, {{"$or": [{{"resources.name"'
        + f': "{tmp_resource_names}"}}]}}]}}'
    )


def file_list_2_search_query(file_list):
    """generates a RQL search query to find listed files

    We get a list of files (with absolute path) and create a RQL search
    query.

    :param file_list: list of files with absolute path
    :type file_list: list of str
    :returns: search query to find the files listed in file_list
    :rtype: str
    """

    if not isinstance(file_list, list):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'file_list' has to be a list of strings but has type "
            + f"{type(file_list).__name__}"
        )

    if len(file_list) == 0:
        return None

    # disassemble the file list;
    #   - split directory and filename
    #   - create dictionary with directories as keys and
    #       lists of filenames as values
    file_dict = dict()
    for iP in file_list:
        i_dir = os.path.dirname(iP)
        i_file = os.path.basename(iP)

        tmp_list = file_dict.get(i_dir, [])
        # append filename if it is not already in the list
        if i_file not in tmp_list:
            tmp_list.append(i_file)
            file_dict[i_dir] = tmp_list

    # construct search query
    #  Differentiate between a situation where all files are in one
    #  directory/namespace in StrongLink (if ...) or in more than
    #  one (else ...).
    if len(file_dict) == 1:
        tmp_dir = list(file_dict.keys())[0]
        return _filename_list_2_search_query(tmp_dir, file_dict[tmp_dir])
    else:
        tmp_query = '{"$or":['
        for tmp_dir, tmp_fn in file_dict.items():
            tmp_query = tmp_query + _filename_list_2_search_query(tmp_dir, tmp_fn)
            tmp_query = tmp_query + ","
        tmp_query = tmp_query[:-1] + "]}"
        return tmp_query


def _parse_year_month_day_2_date(df: pd.DataFrame) -> pd.DataFrame:
    """parse the year, month, day columns into datetime object column."""
    if not isinstance(df, pd.DataFrame):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'df' has to be a 'pandas.DataFrame' but has type "
            + f"{type(df).__name__}"
        )
    if "day" in df.columns and "month" in df.columns and "year" in df.columns:
        df["month"] = df.apply(
            lambda x: list(calendar.month_abbr).index(x.month), axis=1
        )
        dates = pd.to_datetime(dict(year=df.year, month=df.month, day=df.day))
        # move date column to where day column was
        df.insert(loc=df.columns.get_loc("day"), column="date", value=dates)
        return df.drop(columns=["year", "day", "month"])
    else:
        return df


def _parse_iso_str_timestamp_2_datetime(df: pd.DataFrame) -> pd.DataFrame:
    """parse the str ISO timestamps to datetime"""
    if not isinstance(df, pd.DataFrame):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'df' has to be a 'pandas.DataFrame' but has type "
            + f"{type(df).__name__}"
        )
    for col in df:
        if col.startswith("timestamp"):
            df[col] = pd.to_datetime(df[col])
    return df


def _parse_dates(df: pd.DataFrame) -> pd.DataFrame:
    """parse the year, month, day columns into datetime object column."""
    # try:
    #     import pandas as pd
    # except ModuleNotFoundError:
    #     PySlkException(f"pyslk.{inspect.stack()[0][3]}: functions needs "
    #                    "'pandas' to be installed")
    # map month abbreviations to integers
    if not isinstance(df, pd.DataFrame):
        raise TypeError(
            f"pyslk.{inspect.stack()[0][3]}: argument 'df' has to be a 'pandas.DataFrame' but has type "
            + f"{type(df).__name__}"
        )
    return _parse_iso_str_timestamp_2_datetime(_parse_year_month_day_2_date(df))


def _rows_to_dataframe(
    rows: list, cols: list = PYSLK_DEFAULT_LIST_COLUMNS
) -> pd.DataFrame:
    """convert list of rows to dataframe"""
    return pd.DataFrame(rows, columns=cols)


def _parse_sizes(df: pd.DataFrame) -> pd.DataFrame:
    """parses the filesize column into integer bytes"""
    df["filesize"] = df["filesize"].map(_parse_size)
    return df


def _parse_list(
    list_out: str,
    path_or_id: Union[str, int, Path],
    column_names: list = PYSLK_DEFAULT_LIST_COLUMNS,
    parse_dates: bool = True,
    parse_sizes: bool = True,
    numeric_ids: bool = False,
    full_path: bool = True,
    is_resource_id: bool = False,
) -> pd.DataFrame:
    """Return pandas.DataFrame containing results from search id or GNS path

    Calls 'pyslk.pyslk.slk_list(...)' and parses the return string into a
    pandas.DataFrame. All arguments are copied 1:1 into the 'slk_list' call
    except for 'column_widths'. Assumes six output columns of 'slk list' having
    the widths 12, 12, 12, 9, 4, 4, 5 and 999 (999 => as wide as necessary).

    Note: 'slk list' currently only print the modification date and no
    modification time. The output of this parser might be modified in
    future when modification times are printed as well.

    :param list_out: output of slk list
    :type list_out: str
    :param path_or_id: search id or gns path
    :type path_or_id: str or int or Path
    :param column_names: names of the columns in the pandas.DataFrame
    :type column_names: list
    :param parse_dates: parse day, month and year into a datetime column.
    :type parse_dates: bool
    :param parse_sizes: parse filesize column into bytes integer.
    :type parse_sizes: bool
    :param numeric_ids: show numeric values for user and group, default: False
        (show user and group names)
    :type numeric_ids: bool
    :param full_path: add full filepath to filename column.
    :type full_path: bool
    :param is_resource_id: the last column is a resource id - no resource path
    :type is_resource_id: bool
    :returns: output of 'slk list' parsed into a pandas.DataFrame with eight
        columns (permissions, owner, group, size, day, month, year, filename)
    :rtype: pandas.DataFrame
    """
    n_columns_expected = len(column_names)
    n_columns_reality = len(
        [entry for entry in list_out.split("\n")[0].split(" ") if entry]
    )
    if n_columns_reality != n_columns_expected:
        warnings.warn(
            f"pyslk.{inspect.stack()[0][3]}: based on value of 'column_names' expected {n_columns_expected} columns in "
            + f"raw output but got {n_columns_reality} columns"
        )

    rows = _parse_list_to_rows(
        list_out.split("\n"),
        path_or_id=path_or_id,
        full_path=full_path,
        is_resource_id=is_resource_id,
    )
    # split each row into individual columns
    df = _rows_to_dataframe(rows, cols=column_names)

    if parse_dates:
        try:
            df = _parse_dates(df)
        except Exception as e:
            warnings.warn(
                f"pyslk.{inspect.stack()[0][3]}: date parsing failed with: {repr(e)}"
            )

    if parse_sizes:
        try:
            df = _parse_sizes(df)
        except Exception as e:
            warnings.warn(
                f"pyslk.{inspect.stack()[0][3]}: filesize parsing failed with: {repr(e)}"
            )

    if numeric_ids:
        df.group = df.group.astype(int)
        df.owner = df.owner.astype(int)

    if is_resource_id:
        df.filename = df.filename.astype(int)

    return df


def login_valid(quiet: bool = False, rise_error_on_invalid: bool = False) -> bool:
    """
    Check if a valid login token exists.

    Raises 'PySlkNoValidLoginTokenError' when 'rise_error_on_invalid' is True.

    Code calling this function might raise PySlkNoValidLoginTokenError if this functions returns False.

    :param quiet: print no warnings when the slk config file cannot be properly interpreted
    :type quiet: bool
    :param rise_error_on_invalid: raise a 'PySlkNoValidLoginTokenError' when no valid token is present
    :type rise_error_on_invalid: bool
    :return:
    """
    if not os.path.isfile(SLK_USER_CONFIG):
        message = (
            f"slk user config file with login token does not exist ({SLK_USER_CONFIG}). "
            + "Please run 'slk login' to login and to generate this file."
        )
        if rise_error_on_invalid:
            raise PySlkNoValidLoginTokenError(message)
        if not quiet:
            warnings.warn(message)
        return False
    try:
        with open(SLK_USER_CONFIG) as f:
            config_json = json.load(f)
        if not ("expireDate" in config_json and "sessionKey" in config_json):
            message = (
                f"slk user config file seems to be incomplete ({SLK_USER_CONFIG}). Please run 'slk login' "
                + "to login and to generate this file."
            )
            if rise_error_on_invalid:
                raise PySlkNoValidLoginTokenError(message)
            if not quiet:
                warnings.warn(message)
            return False

        exp_date = convert_expiration_date(config_json["expireDate"])
        if exp_date > datetime.now(exp_date.tzinfo):
            return True
        else:
            if rise_error_on_invalid:
                raise PySlkNoValidLoginTokenError(
                    "slk login token is expired. Please run 'slk login' to login to get a fresh token"
                )
            return False
    except json.decoder.JSONDecodeError:
        message = (
            f"slk user config file with login token could not be interpreted ({SLK_USER_CONFIG}). "
            + "Please run 'slk login' to login and to re-generate file."
        )
        if rise_error_on_invalid:
            raise PySlkNoValidLoginTokenError(message)
        if not quiet:
            warnings.warn(message)
        return False


def convert_expiration_date(exp_date: str) -> datetime:
    """
    interprets a date in a string and converts it into a datetime object

    :param exp_date: expiration date of the format '%a %b %d %H:%M:%S %Z %Y'
    :type exp_date: str
    :return: expiration date converted into a datetime object
    :rtype: datetime
    """
    # new approach after issue in KIX ticket 142464
    out_date = parse(exp_date)
    # this is an approach which will cause issues if no 'locale' is set:
    # if locale.getlocale()[0] == "en_US":
    #     out_date = datetime.strptime(exp_date, "%a %b %d %H:%M:%S %Z %Y")
    # else:
    #     curr_locale = locale.getlocale(locale.LC_TIME)
    #    locale.setlocale(locale.LC_TIME, "en_US.utf-8")
    #    out_date = datetime.strptime(exp_date, "%a %b %d %H:%M:%S %Z %Y")
    #    locale.setlocale(locale.LC_TIME, f"{curr_locale[0]}.{curr_locale[1]}")
    return out_date


def access_local(
    resources: Union[list[str], list[Path], str, Path], mode: int
) -> Union[list[bool], bool]:
    if isinstance(resources, str):
        return access_local(Path(resources), mode)
    elif isinstance(resources, list):
        return [access_local(i, mode) for i in resources]
    elif isinstance(resources, Path):
        return os.access(resources, mode)
    else:
        raise TypeError(
            "Need 'str', path-like object or list of one of both types as input. Got "
            + f"'{type(resources).__name__}'."
        )


def is_search_id(path_or_id: Union[Path, str, int]) -> bool:
    if isinstance(path_or_id, int):
        return True
    if isinstance(path_or_id, Path):
        return False
    if isinstance(path_or_id, str):
        try:
            int(path_or_id)
            return True
        except ValueError:
            return False
    raise TypeError(
        f"'path_or_id' has to be 'str', 'int' or path-like but is of type '{type(path_or_id).__name__}'"
    )


def list_local(
    path: Union[str, int],
    show_hidden: bool = False,
    numeric_ids: bool = False,
    recursive: bool = False,
    parse_dates: bool = True,
    parse_sizes: bool = True,
    full_path: bool = True,
) -> pd.DataFrame:
    """Return pandas.DataFrame containing results from search id or GNS path

    Calls 'pyslk.pyslk.slk_list(...)' and parses the return string into a
    pandas.DataFrame. All arguments are copied 1:1 into the 'slk_list' call
    except for 'column_widths'. Assumes six output columns of 'slk list' having
    the widths 12, 12, 12, 9, 4, 4, 5 and 999 (999 => as wide as necessary).

    Note: 'slk list' currently only print the modification date and no
    modification time. The output of this parser might be modified in
    future when modification times are printed as well.

    :param path: search id or gns path
    :type path: str
    :param show_hidden: show '.' files, default: False (don't show these files)
    :type show_hidden: bool
    :param numeric_ids: show numeric values for user and group, default: False
        (show user and group names)
    :type numeric_ids: bool
    :param recursive: use the -R flag to list recursively, default: False
    :type recursive: bool
    :param parse_dates: parse day, month and year into a datetime column.
    :type parse_dates: bool
    :param parse_sizes: parse filesize column into bytes integer.
    :type parse_sizes: bool
    :param full_path: add full filepath to filename column.
    :type full_path: bool
    :returns: output of 'slk list' parsed into a pandas.DataFrame with eight
        columns (permissions, owner, group, size, day, month, year, filename)
    :rtype: pandas.DataFrame
    """
    # get file list and file stats
    files = [
        Path(root, f)
        for root, dirs, files in os.walk(path)
        for f in files
        if not show_hidden or f[0] != "."
    ]
    stats = [os.stat(f) for f in files]
    # permissions
    modes = {s.st_mode: filemode(s.st_mode) for s in stats}
    df = pd.DataFrame([modes[s.st_mode] for s in stats], columns=["permissions"])
    # owner
    if numeric_ids:
        df["owner"] = [s.st_uid for s in stats]
    else:
        uids = {s.st_uid: getpwuid(s.st_uid).pw_name for s in stats}
        df["owner"] = [uids[s.st_uid] for s in stats]
    # groups
    # owner
    if numeric_ids:
        df["group"] = [s.st_gid for s in stats]
    else:
        gids = {s.st_gid: getgrgid(s.st_gid).gr_name for s in stats}
        df["group"] = [gids[s.st_gid] for s in stats]
    # sizes
    if parse_sizes:
        df["filesize"] = [s.st_size for s in stats]
    else:
        df["filesize"] = [_convert_size(s.st_size, "h")["value"] for s in stats]
    # dates
    mtimes = [datetime.utcfromtimestamp(s.st_mtime) for s in stats]
    df["day"] = [t.strftime("%d") for t in mtimes]
    df["month"] = [t.strftime("%b") for t in mtimes]
    df["year"] = [t.strftime("%Y") for t in mtimes]
    df["time"] = [t.strftime("%H:%M") for t in mtimes]
    if parse_dates:
        df = _parse_dates(df)
    # filename
    if full_path:
        df["filename"] = files
    else:
        df["filename"] = [os.path.basename(f) for f in files]
    # return dataframe
    return df


def construct_dst_from_src(
    src: Union[str, Path, list], dst_gns: Union[str, Path], recursion_depth: int = 0
) -> Optional[Union[str, Path, list]]:
    """Construct destination paths for source resources when they were archived to given destination

    :param src:
        Path, pattern or list of files that should be archived.
    :type src: ``str`` or ``list``
    :param dst_gns:
        Destination directory for archiving.
    :type dst_gns: ``str``
    :param recursion_depth:
        count on how often this function has called itself (stop after 50 recursive calls)
    :type recursion_depth: ``int``
    :returns:
        target resource paths if ``src`` resource(s) are archived to ``dst_gns``
    :rtype: ``str`` or ``list`` or ``Path``

    .. seealso::
        * :py:meth:`~pyslk.archive_raw`
        * :py:meth:`~pyslk.archive`

    """
    if recursion_depth > 9:
        raise RuntimeError(
            "construct_dst_from_src: this function called itself recursively more than 10 times; input was: "
            + f"src = {src}; dst_gns = {dst_gns}"
        )
    if isinstance(src, (str, Path)):
        # here we get a list
        src = glob.glob(os.path.expanduser(os.path.expandvars(src)))
        if len(src) == 1:
            # original src matched one resource => proceed
            src = src[0]
            # next: src is a file
            if os.path.isfile(src):
                return os.path.join(dst_gns, os.path.basename(src))
            # when we arrive here, src is a folder
            dst_gns_full = list()
            for root, subdirs, files in os.walk(src):
                if len(files) > 0:
                    # construct destination paths
                    dst_gns_full.extend([str(dst_gns) + "/" + f for f in files])
                if len(subdirs) > 0:
                    # run this function again for all subdirs
                    dst_gns_full.extend(
                        construct_dst_from_src(
                            [root + "/" + s for s in subdirs],
                            dst_gns,
                            recursion_depth=recursion_depth + 1,
                        )
                    )
            return dst_gns_full
        elif len(src) == 0:
            # no match => does not exist
            warnings.warn(
                "construct_dst_from_src: resource does not exist or bash glob yielded no match; input was: "
                + f"src = {src}; dst_gns = {dst_gns}"
            )
            return None
        else:
            # multiple matches: call this command recursively and return output
            return construct_dst_from_src(
                src, dst_gns, recursion_depth=recursion_depth + 1
            )
    if isinstance(src, list):
        dst_gns_full = list()
        for src in src:
            dst_gns_tmp = construct_dst_from_src(
                src, dst_gns, recursion_depth=recursion_depth + 1
            )
            if isinstance(dst_gns_tmp, list):
                dst_gns_full.extend(dst_gns_tmp)
            elif isinstance(dst_gns_tmp, (str, Path)):
                dst_gns_full.append(dst_gns_tmp)
            elif dst_gns_tmp is None:
                pass
            else:
                raise TypeError(
                    "construct_dst_from_src: wrong return type of function call; expected 'str', 'list' or "
                    f"Path-like but got {type(dst_gns_tmp).__name__}"
                )
                pass
        return dst_gns_full
    raise TypeError(
        "construct_dst_from_src: wrong input type for 'src'; expected 'str', 'list' or Path-like but got "
        f"{type(src).__name__}"
    )



from __future__ import annotations

import inspect
import json
import typing
import warnings
from pathlib import Path
from typing import Union

from pyslk.pyslk_exceptions import PySlkException
from pyslk.raw.jobs_verify import submit_verify_job_raw

from .basic import _is_job_completed, is_job_finished, is_job_successful

__all__ = [
    "SubmittedJobs",
]

from ..constants import MAX_RETRIES_SUBMIT_JOBS_BASIC, MAX_RETRIES_SUBMIT_JOBS_SAVE_MODE


class SubmittedJobs:

    required_keys = [
        "jobs",
        "search_id",
        "source",
        "submitted_all_jobs",
        "timeout",
    ]
    optional_keys = [
        "search_query",
        "n_max_search_results",
        "n_max_verify_jobs",
        "n_verify_jobs",
        "restart_information",
    ]
    required_restart_keys = ["search_id", "start_page", "save_mode", "verbose"]

    def __init__(
        self,
        *,
        submitted_jobs_raw_output: Union[str, None] = None,
        resources: Union[str, Path, list[str], list[Path], None] = None,
        resource_ids: Union[int, str, list[int], list[str], None] = None,
        search_id: Union[int, str, None] = None,
        search_query: Union[str, None] = None,
        json_input_file: Union[str, Path, None] = None,
        recursive: bool = False,
        resume_on_page: int = 1,
        results_per_page: int = -1,
        end_on_page: int = -1,
        save_mode: bool = False,
    ):
        """
        Constructor which submits a verify job

        Expected structure of:

        .. code-block:: json

            {'source': 'resources',
            'search_id': 659962,
            'n_max_search_results': 1342935,
            'n_max_verify_jobs': 27,
            'jobs': [{'job_id': 206839, 'offset': 0, 'limit': 50000},
                {'job_id': 206843, 'offset': 50000, 'limit': 50000},
                {'job_id': 206848, 'offset': 100000, 'limit': 50000},
                {'job_id': 206854, 'offset': 150000, 'limit': 50000},
                {'job_id': 206864, 'offset': 200000, 'limit': 50000},
                {'job_id': 206876, 'offset': 250000, 'limit': 50000},
                {'job_id': 206884, 'offset': 300000, 'limit': 50000},
                {'job_id': 206892, 'offset': 350000, 'limit': 50000},
                {'job_id': 206896, 'offset': 400000, 'limit': 50000},
                {'job_id': 206898, 'offset': 450000, 'limit': 50000},
                {'job_id': 206908, 'offset': 500000, 'limit': 50000},
                {'job_id': 206911, 'offset': 550000, 'limit': 50000},
                {'job_id': 206916, 'offset': 600000, 'limit': 50000},
                {'job_id': 206920, 'offset': 650000, 'limit': 50000},
                {'job_id': 206925, 'offset': 700000, 'limit': 50000},
                {'job_id': 206933, 'offset': 750000, 'limit': 50000},
                {'job_id': 206942, 'offset': 800000, 'limit': 50000},
                {'job_id': 206950, 'offset': 850000, 'limit': 50000},
                {'job_id': 206957, 'offset': 900000, 'limit': 50000},
                {'job_id': 206963, 'offset': 950000, 'limit': 50000},
                {'job_id': 206965, 'offset': 1000000, 'limit': 50000},
                {'job_id': 206978, 'offset': 1050000, 'limit': 50000},
                {'job_id': 206983, 'offset': 1100000, 'limit': 50000},
                {'job_id': 206988, 'offset': 1150000, 'limit': 50000},
                {'job_id': 206993, 'offset': 1200000, 'limit': 50000},
                {'job_id': 206999, 'offset': 1250000, 'limit': 50000},
                {'job_id': 207006, 'offset': 1300000, 'limit': 43000}],
            'n_verify_jobs': 27,
            'submitted_all_jobs': True,
            'timeout': False}

        :param resources: provide a list of resources over which the verify job should be run
        :type resources: `str` or `Path` or `list[str]` or `list[Path]` or `None`
        :param resource_ids: provide a list of ids of resources over which the verify job should be run
        :type resource_ids: `str` or `int` or `list[str]` or `list[int]` or `None`
        :param search_id: provide a search id pointing to files over which the verify job should be run
        :type search_id: `str` or `int` or `None`
        :param search_query: provide a search query to select files over which the verify job should be run
        :type search_query: `str` or `None`
        :param json_input_file: read a JSON file which contains payload of the verify job (admin users only)
        :type json_input_file: `str` or `Path`  or `None`
        :param recursive: iterate namespaces recursively
        :type recursive: `bool`
        :param resume_on_page: resume the command and start with search result page `resume_on_page`; internally, 1000
                search results are on one 'page' and fetched by one request; you do not necessarily have read
                permissions for all of these files
        :type resume_on_page: `int`
        :param results_per_page: number of search results requested per page
        :type results_per_page: `int`
        :param end_on_page: end with search result page `end_on_page`; internally, 1000 search results
                are on one 'page' and fetched by one request; you do not necessarily have read permissions for all of
                these files
        :type end_on_page: `int`
        :param save_mode: save mode suggested to be used in times of many timeouts; please do not regularly use this
                parameter; start one verify job per page of search results instead of one verify job for 50 pages of
                search results
        :type save_mode: `bool`
        """

        # ~~~ set class variables ~~~
        # variables for job output
        self.__jobs__: list[dict] = list()
        self.__n_max_search_results__: int = -1
        self.__n_max_verify_jobs__: int = -1
        self.__search_id__: Union[int, str, None] = search_id
        self.__source__: Union[str, None] = None
        self.__submitted_all_jobs__: bool = False
        self.__timeout__: bool = False
        self.__search_query__: Union[str, None] = search_query
        self.__n_verify_jobs__: int = 0
        self.__restart_information__: Union[dict, None] = None
        # some variables for internal processes
        self.__submitted_any_jobs__: bool = False
        # variables for original job specification
        self.__resources__: Union[str, Path, list[str], list[Path], None] = resources
        self.__resource_ids__: Union[int, str, list[int], list[str], None] = (
            resource_ids
        )
        self.__json_input_file__: Union[str, Path, None] = json_input_file
        self.__search_query__: Union[str, None] = search_query
        self.__recursive__: bool = recursive
        self.__resume_on_page__: int = resume_on_page
        self.__results_per_page__: int = results_per_page
        self.__end_on_page__: int = end_on_page
        self.__save_mode__: bool = save_mode

        # we have four cases:
        # 1. submitted_jobs_raw_output is set
        #      This means that someone ran submit_verify_job_raw() and we just extract the results of the str output
        # 2. submitted_jobs_raw_output is None BUT either resources, resource_ids, search_id or json_input_file is set
        #      This means that we store the input values and return; the user then can run 'submit()' when he/she wants
        # 3. more than one of the above-mentioned arguments is set:
        #      error
        # 4. none of the above-mentioned arguments is set:
        #      error
        number_of_set_arguments: int = (
            (submitted_jobs_raw_output is not None)
            + (resources is not None)
            + (resource_ids is not None)
            + (search_id is not None and ())
        )
        if number_of_set_arguments > 1:
            # case 3
            raise PySlkException(
                f"pyslk.{inspect.stack()[0][3]}: more than one of the following arguments is set but only one is "
                + "allowed: 'submitted_jobs_raw_output', 'resources', 'resource_ids', 'search_id', 'json_input_file'."
            )
        if number_of_set_arguments == 0:
            # case 4
            raise PySlkException(
                f"pyslk.{inspect.stack()[0][3]}: none of the following arguments is set but exactly one is required: "
                + "'submitted_jobs_raw_output', 'resources', 'resource_ids', 'search_id', 'json_input_file'."
            )
        if submitted_jobs_raw_output is not None:
            # case 1
            submitted_jobs_dict: dict = json.loads(submitted_jobs_raw_output)
            missing_keys = [
                key for key in self.required_keys if key not in submitted_jobs_dict
            ]
            if len(missing_keys) > 0:
                raise PySlkException(
                    f"pyslk.{inspect.stack()[0][3]}: a few keys were missing in the input 'submitted_jobs_dict': "
                    + f"{missing_keys}. If you just used the output of 'pyslk.submit_verify_jobs_raw', please contact "
                    + "support@dkrz.de"
                )
            # basic job output
            self.__jobs__ = submitted_jobs_dict["jobs"]
            self.__n_max_search_results__ = submitted_jobs_dict["n_max_search_results"]
            self.__n_max_verify_jobs__ = submitted_jobs_dict["n_max_verify_jobs"]
            self.__search_id__ = submitted_jobs_dict["search_id"]
            self.__source__ = submitted_jobs_dict["source"]
            self.__submitted_all_jobs__ = submitted_jobs_dict["submitted_all_jobs"]
            self.__timeout__ = submitted_jobs_dict["timeout"]
            if "search_query" in submitted_jobs_dict:
                self.__search_query__ = submitted_jobs_dict["search_query"]
            else:
                self.__search_query__ = None
            if "n_verify_jobs" in submitted_jobs_dict:
                self.__n_verify_jobs__ = submitted_jobs_dict["n_verify_jobs"]
            else:
                self.__n_verify_jobs__ = 0
            if "restart_information" in submitted_jobs_dict:
                self.__restart_information__ = submitted_jobs_dict[
                    "restart_information"
                ]
            else:
                self.__restart_information__ = None
            # jobs have been submitted at least once
            self.__submitted_any_jobs__ = True
        else:
            # case 2: nothing to be done
            pass
            # just in case we change the default above: no jobs have been submitted
            self.__submitted_any_jobs__ = False

    def __submit_jobs__(
        self,
        resources: Union[str, Path, list[str], list[Path], None] = None,
        resource_ids: Union[int, str, list[int], list[str], None] = None,
        search_id: Union[int, str, None] = None,
        search_query: Union[str, None] = None,
        json_input_file: Union[str, Path, None] = None,
        recursive: bool = False,
        resume_on_page: int = 1,
        results_per_page: int = -1,
        end_on_page: int = -1,
        save_mode: bool = False,
    ) -> bool:
        """
        internal function: submit verify jobs and store the output internally

        :param resources: provide a list of resources over which the verify job should be run
        :type resources: `str` or `Path` or `list[str]` or `list[Path]` or `None`
        :param resource_ids: provide a list of ids of resources over which the verify job should be run
        :type resource_ids: `str` or `int` or `list[str]` or `list[int]` or `None`
        :param search_id: provide a search id pointing to files over which the verify job should be run
        :type search_id: `str` or `int` or `None`
        :param search_query: provide a search query to select files over which the verify job should be run
        :type search_query: `str` or `None`
        :param json_input_file: read a JSON file which contains payload of the verify job (admin users only)
        :type json_input_file: `str` or `Path`  or `None`
        :param recursive: iterate namespaces recursively
        :type recursive: `bool`
        :param resume_on_page: resume the command and start with search result page `resume_on_page`; internally, 1000
                search results are on one 'page' and fetched by one request; you do not necessarily have read
                permissions for all of these files
        :type resume_on_page: `int`
        :param results_per_page: number of search results requested per page
        :type results_per_page: `int`
        :param end_on_page: end with search result page `end_on_page`; internally, 1000 search results
                are on one 'page' and fetched by one request; you do not necessarily have read permissions for all of
                these files
        :type end_on_page: `int`
        :param save_mode: save mode suggested to be used in times of many timeouts; please do not regularly use this
                parameter; start one verify job per page of search results instead of one verify job for 50 pages of
                search results
        :type save_mode: `bool`
        :return: `True` if all jobs were submitted successfully
        :rtype: `bool`
        """

        output = submit_verify_job_raw(
            resources,
            resource_ids,
            search_id,
            search_query,
            json_input_file,
            recursive,
            resume_on_page,
            results_per_page,
            end_on_page,
            save_mode,
            json=True,
            return_type=2,
        )

        # jobs have been submitted at least once
        self.__submitted_any_jobs__ = True

        # we have an error which is not a timeout
        if output.returncode not in [0, 3]:
            raise PySlkException(
                f"pyslk.{inspect.stack()[0][3]}: received error code {output.returncode}. "
                + "Error message: "
                + f"{output.stdout.decode('utf-8').rstrip()} "
                + f"{output.stderr.decode('utf-8').rstrip()}"
            )

        # warn if timeout occurred; no error => some jobs might have been submitted
        if output.returncode == 3:
            warnings.warn(
                f"pyslk.{inspect.stack()[0][3]}: "
                + (MAX_RETRIES_SUBMIT_JOBS_BASIC + MAX_RETRIES_SUBMIT_JOBS_SAVE_MODE)
                + " timeouts occurred while submitting verify jobs. Please start the remaining verify jobs at "
                + "a later point of time. Restart information is given in the Object returned by this function."
            )

        # extract command output
        submitted_jobs_dict: dict = json.loads(output.stdout.decode("utf-8").rstrip())

        # catch some issues
        if (
            self.__search_id__ is not None
            and self.__search_id__ != submitted_jobs_dict["search_id"]
        ):
            raise PySlkException(
                f"pyslk.{inspect.stack()[0][3]}: both there is a mismatch  " + "don't"
            )
        intersection = set(self.get_job_ids()).intersection(
            set(self.extract_job_ids(submitted_jobs_dict["jobs"]))
        )
        if len(intersection) > 0:
            raise PySlkException(
                f"pyslk.{inspect.stack()[0][3]}: the submitted jobs must have different job ids what they do not have; "
                + "same job ids: {', '.join(intersection)}"
            )

        # these fields are not modified if not None:
        #  * self.__search_id__
        #  * self.__search_query__
        #  * self.__resources__
        #  * self.__resource_ids__
        #  * self.__source__
        #  * self.__n_max_search_results__
        #  * self.__n_max_verify_jobs__
        if self.__search_id__ is None:
            self.__search_id__ = submitted_jobs_dict["search_id"]
        if self.__search_query__ is None and "search_query" in submitted_jobs_dict:
            self.__search_query__ = submitted_jobs_dict["search_query"]
        if self.__resources__ is None and "resources" in submitted_jobs_dict:
            self.__resources__ = submitted_jobs_dict["resources"]
        if self.__resource_ids__ is None and "resource_ids" in submitted_jobs_dict:
            self.__resource_ids__ = submitted_jobs_dict["resource_ids"]
        if self.__source__ is None and "source" in submitted_jobs_dict:
            self.__source__ = submitted_jobs_dict["source"]
        if (
            self.__n_max_search_results__ is None
            or (
                isinstance(self.__n_max_search_results__, int)
                and self.__n_max_search_results__ == -1
            )
        ) and "n_max_search_results" in submitted_jobs_dict:
            self.__n_max_search_results__ = submitted_jobs_dict["n_max_search_results"]
        if (
            self.__n_max_verify_jobs__ == -1
            and "n_max_verify_jobs" in submitted_jobs_dict
        ):
            self.__n_max_verify_jobs__ = submitted_jobs_dict["n_max_verify_jobs"]

        # these fields are overwritten because we assume that 'other' contains more up-to-date information:
        self.__submitted_all_jobs__ = submitted_jobs_dict["submitted_all_jobs"]
        self.__timeout__ = submitted_jobs_dict["timeout"]
        if "restart_information" in submitted_jobs_dict:
            self.__restart_information__ = submitted_jobs_dict["restart_information"]
        # are merged
        if "n_verify_jobs" in submitted_jobs_dict:
            self.__n_verify_jobs__ = (
                self.__n_verify_jobs__ + submitted_jobs_dict["n_verify_jobs"]
            )
        self.__jobs__ = self.__jobs__ + submitted_jobs_dict["jobs"]

        # return value
        return self.__submitted_all_jobs__

    def submit(self, force_save_mode: bool = False):
        """
        submit jobs; fresh start of all jobs and resume interrupted submission

        :param force_save_mode: force to submit jobs in save mode; forcing non-save mode is not possible by this switch
        :type force_save_mode: `bool`
        :return: `True` if all jobs were submitted successfully
        :rtype: `bool`
        """
        # check input parameter
        if not isinstance(force_save_mode, bool):
            raise PySlkException(
                f"pyslk.{inspect.stack()[0][3]}: 'force_save_mode' must be a boolean."
            )

        # all jobs submitted => nothing to do
        if self.__submitted_all_jobs__:
            warnings.warn(
                f"pyslk.{inspect.stack()[0][3]}: all jobs have already been submitted"
            )
            return True
        # else: no or some but not all jobs have been submitted

        # no jobs have yet been submitted
        if not self.__submitted_any_jobs__:
            return self.__submit_jobs__(
                self.__resources__,
                self.__resource_ids__,
                self.__search_id__,
                self.__search_query__,
                self.__json_input_file__,
                self.__recursive__,
                self.__resume_on_page__,
                self.__results_per_page__,
                self.__end_on_page__,
                self.__save_mode__ or force_save_mode,
            )
        # else: some but not all jobs have been submitted

        # no restart information available but necessary
        if not self.has_restart_information():
            raise PySlkException(
                f"pyslk.{inspect.stack()[0][3]}: restart information is not available. Some but not all jobs have "
                + "been submitted. Need restart information to submit more."
            )
        # else: some but not all jobs have been submitted; restart information is available to submit more

        # submit more jobs based on restart information
        return self.__submit_jobs__(
            search_id=self.__restart_information__["search_id"],
            resume_on_page=self.__restart_information__["resume_on_page"],
            results_per_page=self.__restart_information__["results_per_page"],
            end_on_page=self.__restart_information__["end_on_page"],
            save_mode=self.__restart_information__["save_mode"] or force_save_mode,
        )

    def needs_restart(self) -> bool:
        return self.__restart_information__ is not None

    def has_restart_information(self) -> bool:
        return self.__restart_information__ is not None

    def missing_restart_information(self) -> list[str]:
        """
        check whether fields are missing in the available restart information

        this json should have been returned by 'pyslk.submit_verify_jobs_raw'

        .. code-block:: json

            {
                "restart_information": {
                    "search_id": 613746,
                    "start_page": 1,
                    "save_mode": false,
                    "verbose": false
                }
            }
        """
        if not self.has_restart_information():
            raise PySlkException(
                f"pyslk.{inspect.stack()[0][3]}: restart information is not available."
            )
        return [
            key
            for key in self.required_restart_keys
            if key not in self.__restart_information__
        ]

    def get_restart_information(self) -> dict:
        if not self.has_restart_information():
            raise PySlkException(
                f"pyslk.{inspect.stack()[0][3]}: restart information is not available."
            )
        return self.__restart_information__

    def get_job_ids(self) -> list[int]:
        return [job["job_id"] for job in self.__jobs__]

    @staticmethod
    def extract_job_ids(jobs: list[dict]) -> list[int]:
        return [job["job_id"] for job in jobs]

    def get_job_information(self, job_id: int) -> typing.Optional[dict]:
        for job in self.__jobs__:
            if job["job_id"] == job_id:
                return job
        return None

    def get_search_id(self) -> int:
        return self.__search_id__

    def get_source(self) -> typing.Optional[str, Path, list[int]]:
        if self.__source__ is None:
            return None
        return self.__source__

    def get_search_query(self) -> typing.Optional[str]:
        if self.__search_query__ is None:
            return None
        return self.__search_query__

    def finished_successfully(self) -> bool:
        return self.__submitted_all_jobs__

    def had_timeout(self) -> bool:
        return self.__timeout__

    def are_finished_short(self) -> bool:
        for job_id in self.get_job_ids():
            if not is_job_finished(job_id):
                return False
        return True

    def are_finished_details(self) -> dict[int, bool]:
        output = {}
        for job_id in self.get_job_ids():
            output[job_id] = is_job_finished(job_id)
        return output

    def are_successful_short(self) -> bool:
        for job_id in self.get_job_ids():
            if not is_job_successful(job_id):
                return False
        return True

    def are_successful_details(self) -> dict[int, bool]:
        output = {}
        for job_id in self.get_job_ids():
            output[job_id] = is_job_successful(job_id)
        return output

    def are_completed_short(self) -> bool:
        for job_id in self.get_job_ids():
            if not _is_job_completed(job_id):
                return False
        return True

    def are_completed_details(self) -> dict[int, bool]:
        output = {}
        for job_id in self.get_job_ids():
            output[job_id] = _is_job_completed(job_id)
        return output

    def merge(self, other: SubmittedJobs):
        if self.__search_id__ != other.__search_id__:
            raise PySlkException(
                f"pyslk.{inspect.stack()[0][3]}: both SubmittedJobs need to be based on the same search id what they "
                + "don't"
            )
        intersection = set(self.get_job_ids()).intersection(set(other.get_job_ids()))
        if len(intersection) > 0:
            raise PySlkException(
                f"pyslk.{inspect.stack()[0][3]}: the SubmittedJobs must have different job ids what they do not have; "
                + "same job ids: {', '.join(intersection)}"
            )

        # these fields are not modified:
        #  * self.__search_query__
        #  * self.__source__
        # these files are set to -1 because it is hard to guess their values
        #  * self.__n_max_search_results__
        #  * self.__n_max_verify_jobs__
        self.__n_max_search_results__ = -1
        self.__n_max_verify_jobs__ = -1
        # these fields are overwritten because we assume that 'other' contains more up-to-date information:
        self.__submitted_all_jobs__ = other.__submitted_all_jobs__
        self.__timeout__ = other.__timeout__
        self.__restart_information__ = other.__restart_information__
        # are merged
        self.__n_verify_jobs__ = self.__n_verify_jobs__ + other.__n_verify_jobs__
        self.__jobs__ = self.__jobs__ + other.__jobs__
